# /*##########################################################################
#
# Copyright (c) 2014-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""OpenGL Plot backend."""

from __future__ import annotations

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "21/12/2018"

import logging
import weakref

import numpy

from .. import items
from .._utils import FLOAT32_MINPOS
from . import BackendBase
from ... import colors
from ... import qt

from ..._glutils import gl
from ... import _glutils as glu
from . import glutils
from .glutils.PlotImageFile import saveImageToFile
from silx.gui.colors import RGBAColorType

_logger = logging.getLogger(__name__)


# TODO idea: BackendQtMixIn class to share code between mpl and gl
# TODO check if OpenGL is available
# TODO make an off-screen mesa backend

# Content #####################################################################


class _ShapeItem(dict):
    def __init__(
        self,
        x,
        y,
        shape,
        color,
        fill,
        overlay,
        linewidth,
        dashoffset,
        dashpattern,
        gapcolor,
    ):
        super(_ShapeItem, self).__init__()

        if shape not in ("polygon", "rectangle", "line", "vline", "hline", "polylines"):
            raise NotImplementedError("Unsupported shape {0}".format(shape))

        x = numpy.asarray(x)
        y = numpy.asarray(y)

        if shape == "rectangle":
            xMin, xMax = x
            x = numpy.array((xMin, xMin, xMax, xMax))
            yMin, yMax = y
            y = numpy.array((yMin, yMax, yMax, yMin))

        # Ignore fill for polylines to mimic matplotlib
        fill = fill if shape != "polylines" else False

        self.update(
            {
                "shape": shape,
                "color": colors.rgba(color),
                "fill": "hatch" if fill else None,
                "x": x,
                "y": y,
                "linewidth": linewidth,
                "dashoffset": dashoffset,
                "dashpattern": dashpattern,
                "gapcolor": gapcolor,
            }
        )


class _MarkerItem(dict):
    def __init__(
        self,
        x,
        y,
        text,
        color,
        symbol,
        symbolsize,
        linewidth,
        dashoffset,
        dashpattern,
        constraint,
        yaxis,
        font,
        bgcolor,
    ):
        super(_MarkerItem, self).__init__()

        if symbol is None:
            symbol = "+"

        # Apply constraint to provided position
        isConstraint = constraint is not None and x is not None and y is not None
        if isConstraint:
            x, y = constraint(x, y)

        self.update(
            {
                "x": x,
                "y": y,
                "text": text,
                "color": colors.rgba(color),
                "constraint": constraint if isConstraint else None,
                "symbol": symbol,
                "symbolsize": symbolsize,
                "linewidth": linewidth,
                "dashoffset": dashoffset,
                "dashpattern": dashpattern,
                "yaxis": yaxis,
                "font": font,
                "bgcolor": bgcolor,
            }
        )


# shaders #####################################################################

_baseVertShd = """
    attribute vec2 position;
    uniform mat4 matrix;
    uniform bvec2 isLog;

    const float oneOverLog10 = 0.43429448190325176;

    void main(void) {
        vec2 posTransformed = position;
        if (isLog.x) {
            posTransformed.x = oneOverLog10 * log(position.x);
        }
        if (isLog.y) {
            posTransformed.y = oneOverLog10 * log(position.y);
        }
        gl_Position = matrix * vec4(posTransformed, 0.0, 1.0);
    }
    """

_baseFragShd = """
    uniform vec4 color;
    uniform int hatchStep;
    uniform float tickLen;

    void main(void) {
        if (tickLen != 0.) {
            if (mod((gl_FragCoord.x + gl_FragCoord.y) / tickLen, 2.) < 1.) {
                gl_FragColor = color;
            } else {
                discard;
            }
        } else if (hatchStep == 0 ||
            mod(gl_FragCoord.x - gl_FragCoord.y, float(hatchStep)) == 0.) {
            gl_FragColor = color;
        } else {
            discard;
        }
    }
    """

_texVertShd = """
   attribute vec2 position;
   attribute vec2 texCoords;
   uniform mat4 matrix;

   varying vec2 coords;

   void main(void) {
        gl_Position = matrix * vec4(position, 0.0, 1.0);
        coords = texCoords;
   }
   """

_texFragShd = """
    uniform sampler2D tex;

    varying vec2 coords;

    void main(void) {
        gl_FragColor = texture2D(tex, coords);
        gl_FragColor.a = 1.0;
    }
    """

# BackendOpenGL ###############################################################


class BackendOpenGL(BackendBase.BackendBase, glu.OpenGLWidget):
    """OpenGL-based Plot backend.

    WARNINGS:
    Unless stated otherwise, this API is NOT thread-safe and MUST be
    called from the main thread.
    When numpy arrays are passed as arguments to the API (through
    :func:`addCurve` and :func:`addImage`), they are copied only if
    required.
    So, the caller should not modify these arrays afterwards.
    """

    _TEXT_MARKER_PADDING = 4

    def __init__(self, plot, parent=None, f=qt.Qt.Widget):
        glu.OpenGLWidget.__init__(
            self,
            parent,
            alphaBufferSize=8,
            depthBufferSize=0,
            stencilBufferSize=0,
            version=(2, 1),
            f=f,
        )
        BackendBase.BackendBase.__init__(self, plot, parent)

        self._defaultFont: qt.QFont = None
        self.__isOpenGLValid = False

        self._backgroundColor = 1.0, 1.0, 1.0, 1.0
        self._dataBackgroundColor = 1.0, 1.0, 1.0, 1.0

        self.matScreenProj = glutils.mat4Identity()

        self._progBase = glu.Program(_baseVertShd, _baseFragShd, attrib0="position")
        self._progTex = glu.Program(_texVertShd, _texFragShd, attrib0="position")
        self._plotFBOs = weakref.WeakKeyDictionary()

        self._keepDataAspectRatio = False

        self._crosshairCursor = None
        self._mousePosInPixels = None

        self._glGarbageCollector = []

        self._plotFrame = glutils.GLPlotFrame2D(
            foregroundColor=(0.0, 0.0, 0.0, 1.0),
            gridColor=(0.7, 0.7, 0.7, 1.0),
            marginRatios=(0.15, 0.1, 0.1, 0.15),
            font=self.getDefaultFont(),
        )
        self._plotFrame.size = (  # Init size with size int
            int(self.getDevicePixelRatio() * 640),
            int(self.getDevicePixelRatio() * 480),
        )

        self.setAutoFillBackground(False)
        self.setMouseTracking(True)

    # QWidget

    _MOUSE_BTNS = {
        qt.Qt.LeftButton: "left",
        qt.Qt.RightButton: "right",
        qt.Qt.MiddleButton: "middle",
    }

    def sizeHint(self):
        return qt.QSize(8 * 80, 6 * 80)  # Mimic MatplotlibBackend

    def mousePressEvent(self, event):
        if event.button() not in self._MOUSE_BTNS:
            return super(BackendOpenGL, self).mousePressEvent(event)
        x, y = qt.getMouseEventPosition(event)
        self._plot.onMousePress(x, y, self._MOUSE_BTNS[event.button()])
        event.accept()

    def mouseMoveEvent(self, event):
        qtPos = qt.getMouseEventPosition(event)

        previousMousePosInPixels = self._mousePosInPixels
        if qtPos == self._mouseInPlotArea(*qtPos):
            devicePixelRatio = self.getDevicePixelRatio()
            devicePos = qtPos[0] * devicePixelRatio, qtPos[1] * devicePixelRatio
            self._mousePosInPixels = devicePos  # Mouse in plot area
        else:
            self._mousePosInPixels = None  # Mouse outside plot area

        if (
            self._crosshairCursor is not None
            and previousMousePosInPixels != self._mousePosInPixels
        ):
            # Avoid replot when cursor remains outside plot area
            self._plot._setDirtyPlot(overlayOnly=True)

        self._plot.onMouseMove(*qtPos)
        event.accept()

    def mouseReleaseEvent(self, event):
        if event.button() not in self._MOUSE_BTNS:
            return super(BackendOpenGL, self).mouseReleaseEvent(event)
        x, y = qt.getMouseEventPosition(event)
        self._plot.onMouseRelease(x, y, self._MOUSE_BTNS[event.button()])
        event.accept()

    def wheelEvent(self, event):
        delta = event.angleDelta().y()
        angleInDegrees = delta / 8.0
        x, y = qt.getMouseEventPosition(event)
        self._plot.onMouseWheel(x, y, angleInDegrees)
        event.accept()

    def leaveEvent(self, _):
        self._plot.onMouseLeaveWidget()

    # OpenGLWidget API

    def initializeGL(self):
        self.__isOpenGLValid = gl.testGL()
        if not self.__isOpenGLValid:
            return

        gl.glClearStencil(0)

        gl.glEnable(gl.GL_BLEND)
        # gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
        gl.glBlendFuncSeparate(
            gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA, gl.GL_ONE, gl.GL_ONE
        )

        # For lines
        gl.glHint(gl.GL_LINE_SMOOTH_HINT, gl.GL_NICEST)

        # For points
        gl.glEnable(gl.GL_VERTEX_PROGRAM_POINT_SIZE)  # OpenGL 2
        gl.glEnable(gl.GL_POINT_SPRITE)  # OpenGL 2
        # gl.glEnable(gl.GL_PROGRAM_POINT_SIZE)

    def _paintDirectGL(self):
        self._renderPlotAreaGL()
        self._plotFrame.render()
        self._renderOverlayGL()

    def _paintFBOGL(self):
        context = glu.Context.getCurrent()
        plotFBOTex = self._plotFBOs.get(context)
        if self._plot._getDirtyPlot() or self._plotFrame.isDirty or plotFBOTex is None:
            self._plotVertices = (
                # Vertex coordinates
                numpy.array(
                    ((-1.0, -1.0), (1.0, -1.0), (-1.0, 1.0), (1.0, 1.0)),
                    dtype=numpy.float32,
                ),
                # Texture coordinates
                numpy.array(
                    ((0.0, 0.0), (1.0, 0.0), (0.0, 1.0), (1.0, 1.0)),
                    dtype=numpy.float32,
                ),
            )
            if (
                plotFBOTex is None
                or plotFBOTex.shape[1] != self._plotFrame.size[0]
                or plotFBOTex.shape[0] != self._plotFrame.size[1]
            ):
                if plotFBOTex is not None:
                    plotFBOTex.discard()
                plotFBOTex = glu.FramebufferTexture(
                    gl.GL_RGBA,
                    shape=(self._plotFrame.size[1], self._plotFrame.size[0]),
                    minFilter=gl.GL_NEAREST,
                    magFilter=gl.GL_NEAREST,
                    wrap=(gl.GL_CLAMP_TO_EDGE, gl.GL_CLAMP_TO_EDGE),
                )
                self._plotFBOs[context] = plotFBOTex

            with plotFBOTex:
                gl.glClearColor(*self._backgroundColor)
                gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_STENCIL_BUFFER_BIT)
                self._renderPlotAreaGL()
                self._plotFrame.render()

        # Render plot in screen coords
        gl.glViewport(0, 0, self._plotFrame.size[0], self._plotFrame.size[1])

        self._progTex.use()
        texUnit = 0

        gl.glUniform1i(self._progTex.uniforms["tex"], texUnit)
        gl.glUniformMatrix4fv(
            self._progTex.uniforms["matrix"],
            1,
            gl.GL_TRUE,
            glutils.mat4Identity().astype(numpy.float32),
        )

        gl.glEnableVertexAttribArray(self._progTex.attributes["position"])
        gl.glVertexAttribPointer(
            self._progTex.attributes["position"],
            2,
            gl.GL_FLOAT,
            gl.GL_FALSE,
            0,
            self._plotVertices[0],
        )

        gl.glEnableVertexAttribArray(self._progTex.attributes["texCoords"])
        gl.glVertexAttribPointer(
            self._progTex.attributes["texCoords"],
            2,
            gl.GL_FLOAT,
            gl.GL_FALSE,
            0,
            self._plotVertices[1],
        )

        with plotFBOTex.texture:
            gl.glDrawArrays(gl.GL_TRIANGLE_STRIP, 0, len(self._plotVertices[0]))

        self._renderOverlayGL()

    def paintGL(self):
        if not self.__isOpenGLValid:
            return

        plot = self._plotRef()
        if plot is None:
            return

        with plot._paintContext():
            with glu.Context.current(self.context()):
                # Release OpenGL resources
                for item in self._glGarbageCollector:
                    item.discard()
                self._glGarbageCollector = []

                gl.glClearColor(*self._backgroundColor)
                gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_STENCIL_BUFFER_BIT)

                # Check if window is large enough
                if self._plotFrame.plotSize <= (2, 2):
                    return

                # Sync plot frame with window
                self._plotFrame.devicePixelRatio = self.getDevicePixelRatio()
                self._plotFrame.dotsPerInch = self.getDotsPerInch()
                # self._paintDirectGL()
                self._paintFBOGL()

    def _renderItems(self, overlay=False):
        """Render items according to :class:`PlotWidget` order

        Note: Scissor test should already be set.

        :param bool overlay:
            False (the default) to render item that are not overlays.
            True to render items that are overlays.
        """
        # Values that are often used
        plotWidth, plotHeight = self._plotFrame.plotSize
        isXLog = self._plotFrame.xAxis.isLog
        isYLog = self._plotFrame.yAxis.isLog
        isYInverted = self._plotFrame.isYAxisInverted

        # Used by marker rendering
        labels = []
        pixelOffset = 3

        context = glutils.RenderContext(
            isXLog=isXLog,
            isYLog=isYLog,
            dpi=self.getDotsPerInch(),
            plotFrame=self._plotFrame,
        )

        for plotItem in self.getItemsFromBackToFront(
            condition=lambda i: i.isVisible() and i.isOverlay() == overlay
        ):
            if plotItem._backendRenderer is None:
                continue

            item = plotItem._backendRenderer

            if isinstance(item, glutils.GLPlotItem):  # Render data items
                gl.glViewport(
                    self._plotFrame.margins.left,
                    self._plotFrame.margins.bottom,
                    plotWidth,
                    plotHeight,
                )
                # Set matrix
                if item.yaxis == "right":
                    context.matrix = self._plotFrame.transformedDataY2ProjMat
                else:
                    context.matrix = self._plotFrame.transformedDataProjMat
                item.render(context)

            elif isinstance(item, _ShapeItem):  # Render shape items
                gl.glViewport(0, 0, self._plotFrame.size[0], self._plotFrame.size[1])

                if (isXLog and numpy.min(item["x"]) < FLOAT32_MINPOS) or (
                    isYLog and numpy.min(item["y"]) < FLOAT32_MINPOS
                ):
                    # Ignore items <= 0. on log axes
                    continue

                if item["shape"] == "hline":
                    width = self._plotFrame.size[0]
                    _, yPixel = self._plotFrame.dataToPixel(
                        0.5 * sum(self._plotFrame.dataRanges[0]), item["y"], axis="left"
                    )
                    subShapes = [
                        numpy.array(
                            ((0.0, yPixel), (width, yPixel)), dtype=numpy.float32
                        )
                    ]

                elif item["shape"] == "vline":
                    xPixel, _ = self._plotFrame.dataToPixel(
                        item["x"], 0.5 * sum(self._plotFrame.dataRanges[1]), axis="left"
                    )
                    height = self._plotFrame.size[1]
                    subShapes = [
                        numpy.array(
                            ((xPixel, 0), (xPixel, height)), dtype=numpy.float32
                        )
                    ]

                else:
                    # Split sub-shapes at not finite values
                    splits = numpy.nonzero(
                        numpy.logical_not(
                            numpy.logical_and(
                                numpy.isfinite(item["x"]), numpy.isfinite(item["y"])
                            )
                        )
                    )[0]
                    splits = numpy.concatenate(([-1], splits, [len(item["x"])]))
                    subShapes = []
                    for begin, end in zip(splits[:-1] + 1, splits[1:]):
                        if end > begin:
                            subShapes.append(
                                numpy.array(
                                    [
                                        self._plotFrame.dataToPixel(x, y, axis="left")
                                        for (x, y) in zip(
                                            item["x"][begin:end], item["y"][begin:end]
                                        )
                                    ]
                                )
                            )

                for points in subShapes:  # Draw each sub-shape
                    # Draw the fill
                    if item["fill"] is not None and item["shape"] not in (
                        "hline",
                        "vline",
                    ):
                        self._progBase.use()
                        gl.glUniformMatrix4fv(
                            self._progBase.uniforms["matrix"],
                            1,
                            gl.GL_TRUE,
                            self.matScreenProj.astype(numpy.float32),
                        )
                        gl.glUniform2i(self._progBase.uniforms["isLog"], False, False)
                        gl.glUniform1f(self._progBase.uniforms["tickLen"], 0.0)

                        shape2D = glutils.FilledShape2D(
                            points, style=item["fill"], color=item["color"]
                        )
                        shape2D.render(
                            posAttrib=self._progBase.attributes["position"],
                            colorUnif=self._progBase.uniforms["color"],
                            hatchStepUnif=self._progBase.uniforms["hatchStep"],
                        )

                    # Draw the stroke
                    if item["dashpattern"] is not None:
                        if item["shape"] != "polylines":
                            # close the polyline
                            points = numpy.append(
                                points, numpy.atleast_2d(points[0]), axis=0
                            )

                        lines = glutils.GLLines2D(
                            points[:, 0],
                            points[:, 1],
                            color=item["color"],
                            gapColor=item["gapcolor"],
                            width=item["linewidth"],
                            dashOffset=item["dashoffset"],
                            dashPattern=item["dashpattern"],
                        )
                        context.matrix = self.matScreenProj
                        lines.render(context)

            elif isinstance(item, _MarkerItem):
                gl.glViewport(0, 0, self._plotFrame.size[0], self._plotFrame.size[1])

                xCoord, yCoord, yAxis = item["x"], item["y"], item["yaxis"]

                if (isXLog and xCoord is not None and xCoord <= 0) or (
                    isYLog and yCoord is not None and yCoord <= 0
                ):
                    # Do not render markers with negative coords on log axis
                    continue

                color = item["color"]
                bgColor = item["bgcolor"]
                if xCoord is None or yCoord is None:
                    if xCoord is None:  # Horizontal line in data space
                        pixelPos = self._plotFrame.dataToPixel(
                            0.5 * sum(self._plotFrame.dataRanges[0]), yCoord, axis=yAxis
                        )

                        if item["text"] is not None:
                            x = (
                                self._plotFrame.size[0]
                                - self._plotFrame.margins.right
                                - pixelOffset
                            )
                            y = pixelPos[1] - pixelOffset
                            label = glutils.Text2D(
                                item["text"],
                                item["font"],
                                x,
                                y,
                                color=color,
                                bgColor=bgColor,
                                align=glutils.RIGHT,
                                valign=glutils.BOTTOM,
                                devicePixelRatio=self.getDevicePixelRatio(),
                                padding=self._TEXT_MARKER_PADDING,
                            )
                            labels.append(label)

                        width = self._plotFrame.size[0]
                        lines = glutils.GLLines2D(
                            (0, width),
                            (pixelPos[1], pixelPos[1]),
                            color=color,
                            width=item["linewidth"],
                            dashOffset=item["dashoffset"],
                            dashPattern=item["dashpattern"],
                        )
                        context.matrix = self.matScreenProj
                        lines.render(context)

                    else:  # yCoord is None: vertical line in data space
                        yRange = self._plotFrame.dataRanges[1 if yAxis == "left" else 2]
                        pixelPos = self._plotFrame.dataToPixel(
                            xCoord, 0.5 * sum(yRange), axis=yAxis
                        )

                        if item["text"] is not None:
                            x = pixelPos[0] + pixelOffset
                            y = self._plotFrame.margins.top + pixelOffset
                            label = glutils.Text2D(
                                item["text"],
                                item["font"],
                                x,
                                y,
                                color=color,
                                bgColor=bgColor,
                                align=glutils.LEFT,
                                valign=glutils.TOP,
                                devicePixelRatio=self.getDevicePixelRatio(),
                                padding=self._TEXT_MARKER_PADDING,
                            )
                            labels.append(label)

                        height = self._plotFrame.size[1]
                        lines = glutils.GLLines2D(
                            (pixelPos[0], pixelPos[0]),
                            (0, height),
                            color=color,
                            width=item["linewidth"],
                            dashOffset=item["dashoffset"],
                            dashPattern=item["dashpattern"],
                        )
                        context.matrix = self.matScreenProj
                        lines.render(context)

                else:
                    xmin, xmax = self._plot.getXAxis().getLimits()
                    ymin, ymax = self._plot.getYAxis(axis=yAxis).getLimits()
                    if not xmin < xCoord < xmax or not ymin < yCoord < ymax:
                        # Do not render markers outside visible plot area
                        continue
                    pixelPos = self._plotFrame.dataToPixel(xCoord, yCoord, axis=yAxis)

                    if isYInverted:
                        valign = glutils.BOTTOM
                        vPixelOffset = -pixelOffset
                    else:
                        valign = glutils.TOP
                        vPixelOffset = pixelOffset

                    if item["text"] is not None:
                        x = pixelPos[0] + pixelOffset
                        y = pixelPos[1] + vPixelOffset
                        label = glutils.Text2D(
                            item["text"],
                            item["font"],
                            x,
                            y,
                            color=color,
                            bgColor=bgColor,
                            align=glutils.LEFT,
                            valign=valign,
                            devicePixelRatio=self.getDevicePixelRatio(),
                            padding=self._TEXT_MARKER_PADDING,
                        )
                        labels.append(label)

                    # For now simple implementation: using a curve for each marker
                    # Should pack all markers to a single set of points
                    marker = glutils.Points2D(
                        (pixelPos[0],),
                        (pixelPos[1],),
                        marker=item["symbol"],
                        color=color,
                        size=item["symbolsize"],
                    )
                    context.matrix = self.matScreenProj
                    marker.render(context)

            else:
                _logger.error("Unsupported item: %s", str(item))
                continue

        # Render marker labels
        gl.glViewport(0, 0, self._plotFrame.size[0], self._plotFrame.size[1])
        for label in labels:
            label.render(self.matScreenProj, self._plotFrame.dotsPerInch)

    def _renderOverlayGL(self):
        """Render overlay layer: overlay items and crosshair."""
        plotWidth, plotHeight = self._plotFrame.plotSize

        # Scissor to plot area
        gl.glScissor(
            self._plotFrame.margins.left,
            self._plotFrame.margins.bottom,
            plotWidth,
            plotHeight,
        )
        gl.glEnable(gl.GL_SCISSOR_TEST)

        self._renderItems(overlay=True)

        # Render crosshair cursor
        if self._crosshairCursor is not None and self._mousePosInPixels is not None:
            self._progBase.use()
            gl.glUniform2i(self._progBase.uniforms["isLog"], False, False)
            gl.glUniform1f(self._progBase.uniforms["tickLen"], 0.0)
            posAttrib = self._progBase.attributes["position"]
            matrixUnif = self._progBase.uniforms["matrix"]
            colorUnif = self._progBase.uniforms["color"]
            hatchStepUnif = self._progBase.uniforms["hatchStep"]

            gl.glViewport(0, 0, self._plotFrame.size[0], self._plotFrame.size[1])

            gl.glUniformMatrix4fv(
                matrixUnif, 1, gl.GL_TRUE, self.matScreenProj.astype(numpy.float32)
            )

            color, lineWidth = self._crosshairCursor
            gl.glUniform4f(colorUnif, *color)
            gl.glUniform1i(hatchStepUnif, 0)

            xPixel, yPixel = self._mousePosInPixels
            xPixel, yPixel = xPixel + 0.5, yPixel + 0.5
            vertices = numpy.array(
                (
                    (0.0, yPixel),
                    (self._plotFrame.size[0], yPixel),
                    (xPixel, 0.0),
                    (xPixel, self._plotFrame.size[1]),
                ),
                dtype=numpy.float32,
            )

            gl.glEnableVertexAttribArray(posAttrib)
            gl.glVertexAttribPointer(
                posAttrib, 2, gl.GL_FLOAT, gl.GL_FALSE, 0, vertices
            )
            gl.glLineWidth(lineWidth)
            gl.glDrawArrays(gl.GL_LINES, 0, len(vertices))

        gl.glDisable(gl.GL_SCISSOR_TEST)

    def _renderPlotAreaGL(self):
        """Render base layer of plot area.

        It renders the background, grid and items except overlays
        """
        plotWidth, plotHeight = self._plotFrame.plotSize

        gl.glScissor(
            self._plotFrame.margins.left,
            self._plotFrame.margins.bottom,
            plotWidth,
            plotHeight,
        )
        gl.glEnable(gl.GL_SCISSOR_TEST)

        if self._dataBackgroundColor != self._backgroundColor:
            gl.glClearColor(*self._dataBackgroundColor)
            gl.glClear(gl.GL_COLOR_BUFFER_BIT)

        self._plotFrame.renderGrid()

        # Matrix
        trBounds = self._plotFrame.transformedDataRanges
        if trBounds.x[0] != trBounds.x[1] and trBounds.y[0] != trBounds.y[1]:
            # Do rendering of items
            self._renderItems(overlay=False)

        gl.glDisable(gl.GL_SCISSOR_TEST)

    def resizeGL(self, width, height):
        if width == 0 or height == 0:  # Do not resize
            return

        self._plotFrame.size = (
            int(self.getDevicePixelRatio() * width),
            int(self.getDevicePixelRatio() * height),
        )

        self.matScreenProj = glutils.mat4Ortho(
            0, self._plotFrame.size[0], self._plotFrame.size[1], 0, 1, -1
        )

        # Store current ranges
        previousXRange = self.getGraphXLimits()
        previousYRange = self.getGraphYLimits(axis="left")
        previousYRightRange = self.getGraphYLimits(axis="right")

        (xMin, xMax), (yMin, yMax), (y2Min, y2Max) = self._plotFrame.dataRanges
        self.setLimits(xMin, xMax, yMin, yMax, y2Min, y2Max)

        # If plot range has changed, then emit signal
        if previousXRange != self.getGraphXLimits():
            self._plot.getXAxis()._emitLimitsChanged()
        if previousYRange != self.getGraphYLimits(axis="left"):
            self._plot.getYAxis(axis="left")._emitLimitsChanged()
        if previousYRightRange != self.getGraphYLimits(axis="right"):
            self._plot.getYAxis(axis="right")._emitLimitsChanged()

    # Add methods

    @staticmethod
    def _castArrayTo(v):
        """Returns best floating type to cast the array to.

        :param numpy.ndarray v: Array to cast
        :rtype: numpy.dtype
        :raise ValueError: If dtype is not supported
        """
        if numpy.issubdtype(v.dtype, numpy.floating):
            return numpy.float32 if v.itemsize <= 4 else numpy.float64
        elif numpy.issubdtype(v.dtype, numpy.integer):
            return numpy.float32 if v.itemsize <= 2 else numpy.float64
        else:
            raise ValueError("Unsupported data type")

    _DASH_PATTERNS = {
        "": (0.0, None),
        " ": (0.0, None),
        "-": (0.0, ()),
        "--": (0.0, (3.7, 1.6, 3.7, 1.6)),
        "-.": (0.0, (6.4, 1.6, 1, 1.6)),
        ":": (0.0, (1, 1.65, 1, 1.65)),
        None: (0.0, None),
    }
    """Convert from linestyle to (offset, (dash pattern))

    Note: dash pattern internal convention differs from matplotlib:
    - None: no line at all
    - (): "solid" line
    """

    def _lineStyleToDashOffsetPattern(
        self, style
    ) -> tuple[float, tuple[float, float, float, float] | tuple[()] | None]:
        """Convert a linestyle to its corresponding offset and dash pattern"""
        if style is None or isinstance(style, str):
            return self._DASH_PATTERNS[style]

        # (offset, (dash pattern)) case
        offset, pattern = style
        if pattern is None:
            # Convert from matplotlib to internal representation of solid
            pattern = ()
        if len(pattern) == 2:
            pattern = pattern * 2
        return float(offset), tuple(float(v) for v in pattern)

    def addCurve(
        self,
        x,
        y,
        color,
        gapcolor,
        symbol,
        linewidth,
        linestyle,
        yaxis,
        xerror,
        yerror,
        fill,
        alpha,
        symbolsize,
        baseline,
    ):
        for parameter in (
            x,
            y,
            color,
            symbol,
            linewidth,
            linestyle,
            yaxis,
            fill,
            symbolsize,
        ):
            assert parameter is not None
        assert yaxis in ("left", "right")

        # Convert input data
        x = numpy.asarray(x)
        y = numpy.asarray(y)

        # Check if float32 is enough
        if (
            self._castArrayTo(x) is numpy.float32
            and self._castArrayTo(y) is numpy.float32
        ):
            dtype = numpy.float32
        else:
            dtype = numpy.float64

        x = numpy.asarray(x, dtype=dtype, order="C")
        y = numpy.asarray(y, dtype=dtype, order="C")

        # Convert errors to float32
        if xerror is not None:
            xerror = numpy.asarray(xerror, dtype=numpy.float32, order="C")
        if yerror is not None:
            yerror = numpy.asarray(yerror, dtype=numpy.float32, order="C")

        # Handle axes log scale: convert data

        if self._plotFrame.xAxis.isLog:
            logX = numpy.log10(x)

            if xerror is not None:
                # Transform xerror so that
                # log10(x) +/- xerror' = log10(x +/- xerror)
                if hasattr(xerror, "shape") and len(xerror.shape) == 2:
                    xErrorMinus, xErrorPlus = xerror[0], xerror[1]
                else:
                    xErrorMinus, xErrorPlus = xerror, xerror
                with numpy.errstate(divide="ignore", invalid="ignore"):
                    # Ignore divide by zero, invalid value encountered in log10
                    xErrorMinus = logX - numpy.log10(x - xErrorMinus)
                xErrorPlus = numpy.log10(x + xErrorPlus) - logX
                xerror = numpy.array((xErrorMinus, xErrorPlus), dtype=numpy.float32)

            x = logX

        isYLog = (yaxis == "left" and self._plotFrame.yAxis.isLog) or (
            yaxis == "right" and self._plotFrame.y2Axis.isLog
        )

        if isYLog:
            logY = numpy.log10(y)

            if yerror is not None:
                # Transform yerror so that
                # log10(y) +/- yerror' = log10(y +/- yerror)
                if hasattr(yerror, "shape") and len(yerror.shape) == 2:
                    yErrorMinus, yErrorPlus = yerror[0], yerror[1]
                else:
                    yErrorMinus, yErrorPlus = yerror, yerror
                with numpy.errstate(divide="ignore", invalid="ignore"):
                    # Ignore divide by zero, invalid value encountered in log10
                    yErrorMinus = logY - numpy.log10(y - yErrorMinus)
                yErrorPlus = numpy.log10(y + yErrorPlus) - logY
                yerror = numpy.array((yErrorMinus, yErrorPlus), dtype=numpy.float32)

            y = logY

        # TODO check if need more filtering of error (e.g., clip to positive)

        # TODO check and improve this
        if len(color) == 4 and type(color[3]) in [type(1), numpy.uint8, numpy.int8]:
            color = numpy.array(color, dtype=numpy.float32) / 255.0

        if isinstance(color, numpy.ndarray) and color.ndim == 2:
            colorArray = color
            color = None
        else:
            colorArray = None
            color = colors.rgba(color)

        if alpha < 1.0:  # Apply image transparency
            if colorArray is not None and colorArray.shape[1] == 4:
                # multiply alpha channel
                colorArray[:, 3] = colorArray[:, 3] * alpha
            if color is not None:
                color = color[0], color[1], color[2], color[3] * alpha

        fillColor = None
        if fill is True:
            fillColor = color

        dashoffset, dashpattern = self._lineStyleToDashOffsetPattern(linestyle)
        curve = glutils.GLPlotCurve2D(
            x,
            y,
            colorArray,
            xError=xerror,
            yError=yerror,
            lineColor=color,
            lineGapColor=gapcolor,
            lineWidth=linewidth,
            lineDashOffset=dashoffset,
            lineDashPattern=dashpattern,
            marker=symbol,
            markerColor=color,
            markerSize=symbolsize,
            fillColor=fillColor,
            baseline=baseline,
            isYLog=isYLog,
        )
        curve.yaxis = "left" if yaxis is None else yaxis

        if yaxis == "right":
            self._plotFrame.isY2Axis = True

        return curve

    def addImage(self, data, origin, scale, colormap, alpha):
        for parameter in (data, origin, scale):
            assert parameter is not None

        if data.ndim == 2:
            # Ensure array is contiguous and eventually convert its type
            dtypes = [
                dtype
                for dtype in (numpy.float32, numpy.float16, numpy.uint8, numpy.uint16)
                if glu.isSupportedGLType(dtype)
            ]
            if data.dtype in dtypes:
                data = numpy.asarray(data, order="C")
            else:
                _logger.info("addImage: Convert %s data to float32", str(data.dtype))
                data = numpy.array(data, dtype=numpy.float32, order="C")

            normalization = colormap.getNormalization()
            if normalization in glutils.GLPlotColormap.SUPPORTED_NORMALIZATIONS:
                # Fast path applying colormap on the GPU
                cmapRange = colormap.getColormapRange(data=data)
                colormapLut = colormap.getNColors(nbColors=256)
                gamma = colormap.getGammaNormalizationParameter()
                nanColor = colors.rgba(colormap.getNaNColor())

                image = glutils.GLPlotColormap(
                    data,
                    origin,
                    scale,
                    colormapLut,
                    normalization,
                    gamma,
                    cmapRange,
                    alpha,
                    nanColor,
                )

            else:  # Fallback applying colormap on CPU
                rgba = colormap.applyToData(data)
                image = glutils.GLPlotRGBAImage(rgba, origin, scale, alpha)

        elif len(data.shape) == 3:
            # For RGB, RGBA data
            assert data.shape[2] in (3, 4)

            if numpy.issubdtype(data.dtype, numpy.floating):
                data = numpy.asarray(data, dtype=numpy.float32)
            elif data.dtype in [numpy.uint8, numpy.uint16]:
                pass
            elif numpy.issubdtype(data.dtype, numpy.integer):
                data = numpy.asarray(data, dtype=numpy.uint8)
            else:
                raise ValueError("Unsupported data type")

            image = glutils.GLPlotRGBAImage(data, origin, scale, alpha)

        else:
            raise RuntimeError("Unsupported data shape {0}".format(data.shape))

        # TODO is this needed?
        if self._plotFrame.xAxis.isLog and image.xMin <= 0.0:
            raise RuntimeError("Cannot add image with X <= 0 with X axis log scale")
        if self._plotFrame.yAxis.isLog and image.yMin <= 0.0:
            raise RuntimeError("Cannot add image with Y <= 0 with Y axis log scale")

        return image

    def addTriangles(self, x, y, triangles, color, alpha):
        # Handle axes log scale: convert data
        if self._plotFrame.xAxis.isLog:
            x = numpy.log10(x)
        if self._plotFrame.yAxis.isLog:
            y = numpy.log10(y)

        triangles = glutils.GLPlotTriangles(x, y, color, triangles, alpha)

        return triangles

    def addShape(
        self, x, y, shape, color, fill, overlay, linestyle, linewidth, gapcolor
    ):
        x = numpy.asarray(x)
        y = numpy.asarray(y)

        # TODO is this needed?
        if self._plotFrame.xAxis.isLog and x.min() <= 0.0:
            raise RuntimeError("Cannot add item with X <= 0 with X axis log scale")
        if self._plotFrame.yAxis.isLog and y.min() <= 0.0:
            raise RuntimeError("Cannot add item with Y <= 0 with Y axis log scale")

        dashoffset, dashpattern = self._lineStyleToDashOffsetPattern(linestyle)
        return _ShapeItem(
            x,
            y,
            shape,
            color,
            fill,
            overlay,
            linewidth,
            dashoffset,
            dashpattern,
            gapcolor,
        )

    def getDefaultFont(self):
        """Returns the default font, used by raw markers and axes labels"""
        if self._defaultFont is None:
            from matplotlib.font_manager import findfont, FontProperties

            font_filename = findfont(FontProperties(family=["sans-serif"]))
            _logger.debug("Load font from mpl: %s", font_filename)
            id = qt.QFontDatabase.addApplicationFont(font_filename)
            family = qt.QFontDatabase.applicationFontFamilies(id)[0]
            font = qt.QFont(family, 10, qt.QFont.Normal, False)
            font.setStyleStrategy(qt.QFont.PreferAntialias)
            self._defaultFont = font
        return self._defaultFont

    def addMarker(
        self,
        x,
        y,
        text,
        color,
        symbol,
        symbolsize: float,
        linestyle,
        linewidth,
        constraint,
        yaxis,
        font,
        bgcolor: RGBAColorType | None,
    ):
        if font is None:
            font = self.getDefaultFont()

        dashoffset, dashpattern = self._lineStyleToDashOffsetPattern(linestyle)
        return _MarkerItem(
            x,
            y,
            text,
            color,
            symbol,
            symbolsize,
            linewidth,
            dashoffset,
            dashpattern,
            constraint,
            yaxis,
            font,
            bgcolor,
        )

    # Remove methods

    def remove(self, item):
        if isinstance(item, glutils.GLPlotItem):
            if item.yaxis == "right":
                # Check if some curves remains on the right Y axis
                y2AxisItems = (
                    item
                    for item in self._plot.getItems()
                    if isinstance(item, items.YAxisMixIn) and item.getYAxis() == "right"
                )
                self._plotFrame.isY2Axis = next(y2AxisItems, None) is not None

            if item.isInitialized():
                self._glGarbageCollector.append(item)

        elif isinstance(item, (_MarkerItem, _ShapeItem)):
            pass  # No-op

        else:
            _logger.error("Unsupported item: %s", str(item))

    # Interaction methods

    _QT_CURSORS = {
        BackendBase.CURSOR_DEFAULT: qt.Qt.ArrowCursor,
        BackendBase.CURSOR_POINTING: qt.Qt.PointingHandCursor,
        BackendBase.CURSOR_SIZE_HOR: qt.Qt.SizeHorCursor,
        BackendBase.CURSOR_SIZE_VER: qt.Qt.SizeVerCursor,
        BackendBase.CURSOR_SIZE_ALL: qt.Qt.SizeAllCursor,
    }

    def setGraphCursorShape(self, cursor):
        if cursor is None:
            super(BackendOpenGL, self).unsetCursor()
        else:
            cursor = self._QT_CURSORS[cursor]
            super(BackendOpenGL, self).setCursor(qt.QCursor(cursor))

    def setGraphCursor(self, flag, color, linewidth, linestyle):
        if linestyle != "-":
            _logger.warning("BackendOpenGL.setGraphCursor linestyle parameter ignored")

        if flag:
            color = colors.rgba(color)
            crosshairCursor = color, linewidth
        else:
            crosshairCursor = None

        if crosshairCursor != self._crosshairCursor:
            self._crosshairCursor = crosshairCursor

    _PICK_OFFSET = 3  # Offset in pixel used for picking

    def _mouseInPlotArea(self, x, y):
        """Returns closest visible position in the plot.

        This is performed in Qt widget pixel, not device pixel.

        :param float x: X coordinate in Qt widget pixel
        :param float y: Y coordinate in Qt widget pixel
        :return: (x, y) closest point in the plot.
        :rtype: List[float]
        """
        left, top, width, height = self.getPlotBoundsInPixels()
        return (
            numpy.clip(x, left, left + width - 1),  # TODO -1?
            numpy.clip(y, top, top + height - 1),
        )

    def __pickCurves(self, item, x, y):
        """Perform picking on a curve item.

        :param GLPlotCurve2D item:
        :param float x: X position of the mouse in widget coordinates
        :param float y: Y position of the mouse in widget coordinates
        :return: List of indices of picked points or None if not picked
        :rtype: Union[List[int],None]
        """
        offset = self._PICK_OFFSET
        if item.marker is not None:
            # Convert markerSize from points to qt pixels
            qtDpi = self.getDotsPerInch() / self.getDevicePixelRatio()
            size = item.markerSize / 72.0 * qtDpi
            offset = max(size / 2.0, offset)
        if item.lineDashPattern is not None:
            # Convert line width from points to qt pixels
            qtDpi = self.getDotsPerInch() / self.getDevicePixelRatio()
            lineWidth = item.lineWidth / 72.0 * qtDpi
            offset = max(lineWidth / 2.0, offset)

        inAreaPos = self._mouseInPlotArea(x - offset, y - offset)
        dataPos = self._plot.pixelToData(
            inAreaPos[0], inAreaPos[1], axis=item.yaxis, check=True
        )
        if dataPos is None:
            return None
        xPick0, yPick0 = dataPos

        inAreaPos = self._mouseInPlotArea(x + offset, y + offset)
        dataPos = self._plot.pixelToData(
            inAreaPos[0], inAreaPos[1], axis=item.yaxis, check=True
        )
        if dataPos is None:
            return None
        xPick1, yPick1 = dataPos

        if xPick0 < xPick1:
            xPickMin, xPickMax = xPick0, xPick1
        else:
            xPickMin, xPickMax = xPick1, xPick0

        if yPick0 < yPick1:
            yPickMin, yPickMax = yPick0, yPick1
        else:
            yPickMin, yPickMax = yPick1, yPick0

        # Apply log scale if axis is log
        if self._plotFrame.xAxis.isLog:
            xPickMin = numpy.log10(xPickMin)
            xPickMax = numpy.log10(xPickMax)

        if (item.yaxis == "left" and self._plotFrame.yAxis.isLog) or (
            item.yaxis == "right" and self._plotFrame.y2Axis.isLog
        ):
            yPickMin = numpy.log10(yPickMin)
            yPickMax = numpy.log10(yPickMax)

        return item.pick(xPickMin, yPickMin, xPickMax, yPickMax)

    def pickItem(self, x, y, item):
        # Picking is performed in Qt widget pixels not device pixels
        dataPos = self._plot.pixelToData(x, y, axis="left", check=True)
        if dataPos is None:
            return None  # Outside plot area

        if item is None:
            _logger.error("No item provided for picking")
            return None

        # Pick markers
        if isinstance(item, _MarkerItem):
            yaxis = item["yaxis"]
            pixelPos = self._plot.dataToPixel(
                item["x"], item["y"], axis=yaxis, check=False
            )
            if pixelPos is None:
                return None  # negative coord on a log axis

            if item["x"] is None:  # Horizontal line
                pt1 = self._plot.pixelToData(
                    x, y - self._PICK_OFFSET, axis=yaxis, check=False
                )
                pt2 = self._plot.pixelToData(
                    x, y + self._PICK_OFFSET, axis=yaxis, check=False
                )
                isPicked = min(pt1[1], pt2[1]) <= item["y"] <= max(pt1[1], pt2[1])

            elif item["y"] is None:  # Vertical line
                pt1 = self._plot.pixelToData(
                    x - self._PICK_OFFSET, y, axis=yaxis, check=False
                )
                pt2 = self._plot.pixelToData(
                    x + self._PICK_OFFSET, y, axis=yaxis, check=False
                )
                isPicked = min(pt1[0], pt2[0]) <= item["x"] <= max(pt1[0], pt2[0])

            else:
                isPicked = (
                    numpy.fabs(x - pixelPos[0]) <= self._PICK_OFFSET
                    and numpy.fabs(y - pixelPos[1]) <= self._PICK_OFFSET
                )

            return (0,) if isPicked else None

        # Pick image, curve, triangles
        elif isinstance(item, glutils.GLPlotItem):
            if isinstance(item, glutils.GLPlotCurve2D):
                return self.__pickCurves(item, x, y)
            else:
                return item.pick(*dataPos)  # Might be None

    # Update curve

    def setCurveColor(self, curve, color):
        pass  # TODO

    # Misc.

    def getWidgetHandle(self):
        return self

    def postRedisplay(self):
        self.update()

    def replot(self):
        self.update()  # async redraw

    def saveGraph(self, fileName, fileFormat, dpi):
        if dpi is not None:
            _logger.warning("saveGraph ignores dpi parameter")

        if fileFormat not in ["png", "ppm", "svg", "tif", "tiff"]:
            raise NotImplementedError("Unsupported format: %s" % fileFormat)

        if not self.isValid():
            _logger.error("OpenGL 2.1 not available, cannot save OpenGL image")
            width, height = self._plotFrame.size
            data = numpy.zeros((height, width, 3), dtype=numpy.uint8)
        else:
            self.makeCurrent()

            data = numpy.empty(
                (self._plotFrame.size[1], self._plotFrame.size[0], 3),
                dtype=numpy.uint8,
                order="C",
            )

            context = self.context()
            framebufferTexture = self._plotFBOs.get(context)
            if framebufferTexture is None:
                # Fallback, supports direct rendering mode: _paintDirectGL
                # might have issues as it can read on-screen framebuffer
                fboName = self.defaultFramebufferObject()
                width, height = self._plotFrame.size
            else:
                fboName = framebufferTexture.name
                height, width = framebufferTexture.shape

            previousFramebuffer = gl.glGetInteger(gl.GL_FRAMEBUFFER_BINDING)
            gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, fboName)
            gl.glPixelStorei(gl.GL_PACK_ALIGNMENT, 1)
            gl.glReadPixels(0, 0, width, height, gl.GL_RGB, gl.GL_UNSIGNED_BYTE, data)
            gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, previousFramebuffer)

            # glReadPixels gives bottom to top,
            # while images are stored as top to bottom
            data = numpy.flipud(data)

        # fileName is either a file-like object or a str
        saveImageToFile(data, fileName, fileFormat)

    # Graph labels

    def setGraphTitle(self, title):
        self._plotFrame.title = title

    def setGraphXLabel(self, label):
        self._plotFrame.xAxis.title = label

    def setGraphYLabel(self, label, axis):
        if axis == "left":
            self._plotFrame.yAxis.title = label
        else:  # right axis
            self._plotFrame.y2Axis.title = label

    # Graph limits

    def _setDataRanges(self, xlim=None, ylim=None, y2lim=None):
        """Set the visible range of data in the plot frame.

        This clips the ranges to possible values (takes care of float32
        range + positive range for log).
        This also takes care of non-orthogonal axes.

        This should be moved to PlotFrame.
        """
        # Update axes range with a clipped range if too wide
        self._plotFrame.setDataRanges(xlim, ylim, y2lim)

    def _ensureAspectRatio(self, keepDim=None):
        """Update plot bounds in order to keep aspect ratio.

        Warning: keepDim on right Y axis is not implemented !

        :param str keepDim: The dimension to maintain: 'x', 'y' or None.
            If None (the default), the dimension with the largest range.
        """
        plotWidth, plotHeight = self._plotFrame.plotSize
        if plotWidth <= 2 or plotHeight <= 2:
            return

        if keepDim is None:
            ranges = self._plot.getDataRange()
            if (
                ranges.y is not None
                and ranges.x is not None
                and (ranges.y[1] - ranges.y[0]) != 0.0
            ):
                dataRatio = (ranges.x[1] - ranges.x[0]) / float(
                    ranges.y[1] - ranges.y[0]
                )
                plotRatio = plotWidth / float(plotHeight)  # Test != 0 before

                keepDim = "x" if dataRatio > plotRatio else "y"
            else:  # Limit case
                keepDim = "x"

        (xMin, xMax), (yMin, yMax), (y2Min, y2Max) = self._plotFrame.dataRanges
        if keepDim == "y":
            dataW = (yMax - yMin) * plotWidth / float(plotHeight)
            xCenter = 0.5 * (xMin + xMax)
            xMin = xCenter - 0.5 * dataW
            xMax = xCenter + 0.5 * dataW
        elif keepDim == "x":
            dataH = (xMax - xMin) * plotHeight / float(plotWidth)
            yCenter = 0.5 * (yMin + yMax)
            yMin = yCenter - 0.5 * dataH
            yMax = yCenter + 0.5 * dataH
            y2Center = 0.5 * (y2Min + y2Max)
            y2Min = y2Center - 0.5 * dataH
            y2Max = y2Center + 0.5 * dataH
        else:
            raise RuntimeError("Unsupported dimension to keep: %s" % keepDim)

        # Update plot frame bounds
        self._setDataRanges(xlim=(xMin, xMax), ylim=(yMin, yMax), y2lim=(y2Min, y2Max))

    def _setPlotBounds(self, xRange=None, yRange=None, y2Range=None, keepDim=None):
        # Update axes range with a clipped range if too wide
        self._setDataRanges(xlim=xRange, ylim=yRange, y2lim=y2Range)

        # Keep data aspect ratio
        if self.isKeepDataAspectRatio():
            self._ensureAspectRatio(keepDim)

    def setLimits(self, xmin, xmax, ymin, ymax, y2min=None, y2max=None):
        assert xmin < xmax
        assert ymin < ymax

        if y2min is None or y2max is None:
            y2Range = None
        else:
            assert y2min < y2max
            y2Range = y2min, y2max
        self._setPlotBounds((xmin, xmax), (ymin, ymax), y2Range)

    def getGraphXLimits(self):
        return self._plotFrame.dataRanges.x

    def setGraphXLimits(self, xmin, xmax):
        assert xmin < xmax
        self._setPlotBounds(xRange=(xmin, xmax), keepDim="x")

    def getGraphYLimits(self, axis):
        assert axis in ("left", "right")
        if axis == "left":
            return self._plotFrame.dataRanges.y
        else:
            return self._plotFrame.dataRanges.y2

    def setGraphYLimits(self, ymin, ymax, axis):
        assert ymin < ymax
        assert axis in ("left", "right")

        if axis == "left":
            self._setPlotBounds(yRange=(ymin, ymax), keepDim="y")
        else:
            self._setPlotBounds(y2Range=(ymin, ymax), keepDim="y")

    # Graph axes

    def getXAxisTimeZone(self):
        return self._plotFrame.xAxis.timeZone

    def setXAxisTimeZone(self, tz):
        self._plotFrame.xAxis.timeZone = tz

    def isXAxisTimeSeries(self):
        return self._plotFrame.xAxis.isTimeSeries

    def setXAxisTimeSeries(self, isTimeSeries):
        self._plotFrame.xAxis.isTimeSeries = isTimeSeries

    def setXAxisLogarithmic(self, flag):
        if flag != self._plotFrame.xAxis.isLog:
            if flag and self._keepDataAspectRatio:
                _logger.warning("KeepDataAspectRatio is ignored with log axes")

            self._plotFrame.xAxis.isLog = flag

    def setYAxisLogarithmic(self, flag):
        if flag != self._plotFrame.yAxis.isLog or flag != self._plotFrame.y2Axis.isLog:
            if flag and self._keepDataAspectRatio:
                _logger.warning("KeepDataAspectRatio is ignored with log axes")

            self._plotFrame.yAxis.isLog = flag
            self._plotFrame.y2Axis.isLog = flag

    def setYAxisInverted(self, flag):
        if flag != self._plotFrame.isYAxisInverted:
            self._plotFrame.isYAxisInverted = flag

    def isYAxisInverted(self):
        return self._plotFrame.isYAxisInverted

    def isYRightAxisVisible(self):
        return self._plotFrame.isY2Axis

    def isKeepDataAspectRatio(self):
        if self._plotFrame.xAxis.isLog or self._plotFrame.yAxis.isLog:
            return False
        else:
            return self._keepDataAspectRatio

    def setKeepDataAspectRatio(self, flag):
        if flag and (self._plotFrame.xAxis.isLog or self._plotFrame.yAxis.isLog):
            _logger.warning("KeepDataAspectRatio is ignored with log axes")

        self._keepDataAspectRatio = flag

    def setGraphGrid(self, which):
        assert which in (None, "major", "both")
        self._plotFrame.grid = which is not None  # TODO True grid support

    # Data <-> Pixel coordinates conversion

    def dataToPixel(self, x, y, axis):
        result = self._plotFrame.dataToPixel(x, y, axis)
        if result is None:
            return None
        else:
            devicePixelRatio = self.getDevicePixelRatio()
            return tuple(value / devicePixelRatio for value in result)

    def pixelToData(self, x, y, axis):
        devicePixelRatio = self.getDevicePixelRatio()
        return self._plotFrame.pixelToData(
            x * devicePixelRatio, y * devicePixelRatio, axis
        )

    def getPlotBoundsInPixels(self):
        devicePixelRatio = self.getDevicePixelRatio()
        return tuple(
            int(value / devicePixelRatio)
            for value in self._plotFrame.plotOrigin + self._plotFrame.plotSize
        )

    def setAxesMargins(self, left: float, top: float, right: float, bottom: float):
        self._plotFrame.marginRatios = left, top, right, bottom

    def setForegroundColors(self, foregroundColor, gridColor):
        self._plotFrame.foregroundColor = foregroundColor
        self._plotFrame.gridColor = gridColor

    def setBackgroundColors(self, backgroundColor, dataBackgroundColor):
        self._backgroundColor = backgroundColor
        self._dataBackgroundColor = dataBackgroundColor



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""Tests for url module"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "29/01/2018"


import pytest
from ..url import DataUrl


def assert_url(url, expected):
    assert url.is_valid() == expected[0]
    assert url.is_absolute() == expected[1]
    assert url.scheme() == expected[2]
    assert url.file_path() == expected[3]
    assert url.data_path() == expected[4]
    assert url.data_slice() == expected[5]


def test_fabio_absolute():
    url = DataUrl("fabio:///data/image.edf?slice=2")
    expected = [True, True, "fabio", "/data/image.edf", None, (2,)]
    assert_url(url, expected)


def test_fabio_absolute_windows():
    url = DataUrl("fabio:///C:/data/image.edf?slice=2")
    expected = [True, True, "fabio", "C:/data/image.edf", None, (2,)]
    assert_url(url, expected)


def test_silx_absolute():
    url = DataUrl("silx:///data/image.h5?path=/data/dataset&slice=1,5")
    expected = [True, True, "silx", "/data/image.h5", "/data/dataset", (1, 5)]
    assert_url(url, expected)


def test_commandline_shell_separator():
    url = DataUrl("silx:///data/image.h5::path=/data/dataset&slice=1,5")
    expected = [True, True, "silx", "/data/image.h5", "/data/dataset", (1, 5)]
    assert_url(url, expected)


def test_silx_absolute2():
    url = DataUrl("silx:///data/image.edf?/scan_0/detector/data")
    expected = [True, True, "silx", "/data/image.edf", "/scan_0/detector/data", None]
    assert_url(url, expected)


def test_silx_absolute_windows():
    url = DataUrl("silx:///C:/data/image.h5?/scan_0/detector/data")
    expected = [True, True, "silx", "C:/data/image.h5", "/scan_0/detector/data", None]
    assert_url(url, expected)


def test_silx_relative():
    url = DataUrl("silx:./image.h5")
    expected = [True, False, "silx", "./image.h5", None, None]
    assert_url(url, expected)


def test_fabio_relative():
    url = DataUrl("fabio:./image.edf")
    expected = [True, False, "fabio", "./image.edf", None, None]
    assert_url(url, expected)


def test_silx_relative2():
    url = DataUrl("silx:image.h5")
    expected = [True, False, "silx", "image.h5", None, None]
    assert_url(url, expected)


def test_fabio_relative2():
    url = DataUrl("fabio:image.edf")
    expected = [True, False, "fabio", "image.edf", None, None]
    assert_url(url, expected)


def test_file_relative():
    url = DataUrl("image.edf")
    expected = [True, False, None, "image.edf", None, None]
    assert_url(url, expected)


def test_file_relative2():
    url = DataUrl("./foo/bar/image.edf")
    expected = [True, False, None, "./foo/bar/image.edf", None, None]
    assert_url(url, expected)


def test_file_relative3():
    url = DataUrl("foo/bar/image.edf")
    expected = [True, False, None, "foo/bar/image.edf", None, None]
    assert_url(url, expected)


def test_file_absolute():
    url = DataUrl("/data/image.edf")
    expected = [True, True, None, "/data/image.edf", None, None]
    assert_url(url, expected)


def test_file_absolute_windows():
    url = DataUrl("C:/data/image.edf")
    expected = [True, True, None, "C:/data/image.edf", None, None]
    assert_url(url, expected)


def test_absolute_with_path():
    url = DataUrl("/foo/foobar.h5?/foo/bar")
    expected = [True, True, None, "/foo/foobar.h5", "/foo/bar", None]
    assert_url(url, expected)


def test_windows_file_data_slice():
    url = DataUrl("C:/foo/foobar.h5?path=/foo/bar&slice=5,1")
    expected = [True, True, None, "C:/foo/foobar.h5", "/foo/bar", (5, 1)]
    assert_url(url, expected)


def test_scheme_file_data_slice():
    url = DataUrl("silx:/foo/foobar.h5?path=/foo/bar&slice=5,1")
    expected = [True, True, "silx", "/foo/foobar.h5", "/foo/bar", (5, 1)]
    assert_url(url, expected)


def test_scheme_windows_file_data_slice():
    url = DataUrl("silx:C:/foo/foobar.h5?path=/foo/bar&slice=5,1")
    expected = [True, True, "silx", "C:/foo/foobar.h5", "/foo/bar", (5, 1)]
    assert_url(url, expected)


def test_empty():
    url = DataUrl("")
    expected = [False, False, None, "", None, None]
    assert_url(url, expected)


def test_unknown_scheme():
    url = DataUrl("foo:/foo/foobar.h5?path=/foo/bar&slice=5,1")
    expected = [False, True, "foo", "/foo/foobar.h5", "/foo/bar", (5, 1)]
    assert_url(url, expected)


def test_slice():
    url = DataUrl("/a.h5?path=/b&slice=5,1")
    expected = [True, True, None, "/a.h5", "/b", (5, 1)]
    assert_url(url, expected)


def test_slice2():
    url = DataUrl("/a.h5?path=/b&slice=2:5")
    expected = [True, True, None, "/a.h5", "/b", (slice(2, 5),)]
    assert_url(url, expected)


def test_slice3():
    url = DataUrl("/a.h5?path=/b&slice=::2")
    expected = [True, True, None, "/a.h5", "/b", (slice(None, None, 2),)]
    assert_url(url, expected)


def test_slice_ellipsis():
    url = DataUrl("/a.h5?path=/b&slice=...")
    expected = [True, True, None, "/a.h5", "/b", (Ellipsis,)]
    assert_url(url, expected)


def test_slice_slicing():
    url = DataUrl("/a.h5?path=/b&slice=:")
    expected = [True, True, None, "/a.h5", "/b", (slice(None),)]
    assert_url(url, expected)


def test_slice_missing_element():
    url = DataUrl("/a.h5?path=/b&slice=5,,1")
    expected = [False, True, None, "/a.h5", "/b", None]
    assert_url(url, expected)


def test_slice_no_elements():
    url = DataUrl("/a.h5?path=/b&slice=")
    expected = [False, True, None, "/a.h5", "/b", None]
    assert_url(url, expected)


def test_create_relative_url():
    url = DataUrl(scheme="silx", file_path="./foo.h5", data_path="/", data_slice=(5, 1))
    assert not url.is_absolute()
    url2 = DataUrl(url.path())
    assert url == url2


def test_create_absolute_url():
    url = DataUrl(scheme="silx", file_path="/foo.h5", data_path="/", data_slice=(5, 1))
    url2 = DataUrl(url.path())
    assert url == url2


def test_create_absolute_windows_url():
    url = DataUrl(
        scheme="silx", file_path="C:/foo.h5", data_path="/", data_slice=(5, 1)
    )
    url2 = DataUrl(url.path())
    assert url == url2


def test_create_slice_url():
    url = DataUrl(
        scheme="silx",
        file_path="/foo.h5",
        data_path="/",
        data_slice=(5, 1, Ellipsis, slice(None)),
    )
    url2 = DataUrl(url.path())
    assert url == url2


def test_wrong_url():
    url = DataUrl(scheme="silx", file_path="/foo.h5", data_slice=(5, 1))
    assert not url.is_valid()


@pytest.mark.parametrize(
    "data",
    [
        (1, "silx:///foo.h5?slice=1"),
        ((1,), "silx:///foo.h5?slice=1"),
        (slice(None), "silx:///foo.h5?slice=:"),
        (slice(1, None), "silx:///foo.h5?slice=1:"),
        (slice(None, -2), "silx:///foo.h5?slice=:-2"),
        (slice(1, None, 3), "silx:///foo.h5?slice=1::3"),
        (slice(None, 2, 3), "silx:///foo.h5?slice=:2:3"),
        (slice(None, None, 3), "silx:///foo.h5?slice=::3"),
        (slice(1, 2, 3), "silx:///foo.h5?slice=1:2:3"),
        ((1, slice(1, 2)), "silx:///foo.h5?slice=1,1:2"),
    ],
)
def test_path_creation(data):
    """make sure the construction of path succeed and that we can
    recreate a DataUrl from a path"""
    data_slice, expected_path = data
    url = DataUrl(scheme="silx", file_path="/foo.h5", data_slice=data_slice)
    path = url.path()
    DataUrl(path=path)
    assert path == expected_path


def test_file_path_none():
    """
    make sure a file path can be None
    """
    url = DataUrl(scheme="silx", file_path=None, data_path="/path/to/data")
    assert url.file_path() is None
    assert url.scheme() == "silx"
    assert url.data_path() == "/path/to/data"


def test_data_path_none():
    """
    make sure a data path can be None
    """
    url = DataUrl(scheme="silx", file_path="my_file.hdf5", data_path=None)
    assert url.file_path() == "my_file.hdf5"
    assert url.scheme() == "silx"
    assert url.data_path() is None


def test_scheme_none():
    """
    make sure a scheme can be None
    """
    url = DataUrl(scheme=None, file_path="my_file.hdf5", data_path="/path/to/data")
    assert url.file_path() == "my_file.hdf5"
    assert url.scheme() is None
    assert url.data_path() == "/path/to/data"


def test_http_scheme():
    url = DataUrl("http://hsds-server.tld/home/file")
    expected = [True, False, "http", "hsds-server.tld/home/file", None, None]
    assert_url(url, expected)
    assert url.path() == "http://hsds-server.tld/home/file"

def test_http_scheme_with_port():
    url = DataUrl("http://hsds-server.tld:8080/home/file")
    expected = [True, False, "http", "hsds-server.tld:8080/home/file", None, None]
    assert_url(url, expected)
    assert url.path() == "http://hsds-server.tld:8080/home/file"

def test_http_with_path():
    url = DataUrl("http://hsds-server.tld/home/file?/foo/bar")
    expected = [True, False, "http", "hsds-server.tld/home/file", "/foo/bar", None]
    assert_url(url, expected)



# /*##########################################################################
#
# Copyright (c) 2018-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This provides profile ROIs.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "09/06/2023"


import numpy

from silx.gui.plot.tools.profile import rois
from silx.gui.plot.tools.profile import core
from .core import _CompareImageItem


COLOR_A = "C0"
COLOR_B = "C8"


class ProfileImageLineROI(rois.ProfileImageLineROI):
    """ROI for a compare image profile between 2 points.

    The X profile of this ROI is the projection into one of the x/y axes,
    using its scale and its orientation.
    """

    def computeProfile(self, item):
        if not isinstance(item, _CompareImageItem):
            raise TypeError("Unexpected class %s" % type(item))

        origin = item.getOrigin()
        scale = item.getScale()
        method = self.getProfileMethod()
        lineWidth = self.getProfileLineWidth()
        roiInfo = self._getRoiInfo()

        def createProfile2(currentData):
            coords, profile, _area, profileName, xLabel = core.createProfile(
                roiInfo=roiInfo,
                currentData=currentData,
                origin=origin,
                scale=scale,
                lineWidth=lineWidth,
                method=method,
            )
            return coords, profile, profileName, xLabel

        currentData1 = item.getImageData1()
        currentData2 = item.getImageData2()

        yLabel = "%s" % str(method).capitalize()
        coords, profile1, title, xLabel = createProfile2(currentData1)
        title = title + "; width = %d" % lineWidth
        _coords, profile2, _title, _xLabel = createProfile2(currentData2)

        profile1.shape = -1
        profile2.shape = -1

        title = title.format(xlabel="width", ylabel="height")
        xLabel = xLabel.format(xlabel="width", ylabel="height")
        yLabel = yLabel.format(xlabel="width", ylabel="height")

        data = core.CurvesProfileData(
            coords=coords,
            profiles=[
                core.CurveProfileDesc(profile1, color=COLOR_A, name="profileA"),
                core.CurveProfileDesc(profile2, color=COLOR_B, name="profileB"),
            ],
            title=title,
            xLabel=xLabel,
            yLabel=yLabel,
        )
        return data


class ProfileImageDirectedLineROI(rois.ProfileImageDirectedLineROI):
    """ROI for a compare image profile between 2 points.

    The X profile of the line is displayed projected into the line itself,
    using its scale and its orientation. It's the distance from the origin.
    """

    def computeProfile(self, item):
        if not isinstance(item, _CompareImageItem):
            raise TypeError("Unexpected class %s" % type(item))

        from silx.image.bilinear import BilinearImage

        origin = item.getOrigin()
        scale = item.getScale()
        method = self.getProfileMethod()
        lineWidth = self.getProfileLineWidth()

        roiInfo = self._getRoiInfo()
        roiStart, roiEnd, _lineProjectionMode = roiInfo

        startPt = (
            (roiStart[1] - origin[1]) / scale[1],
            (roiStart[0] - origin[0]) / scale[0],
        )
        endPt = ((roiEnd[1] - origin[1]) / scale[1], (roiEnd[0] - origin[0]) / scale[0])

        if numpy.array_equal(startPt, endPt):
            return None

        def computeProfile(data):
            bilinear = BilinearImage(data)
            profile = bilinear.profile_line(
                (startPt[0] - 0.5, startPt[1] - 0.5),
                (endPt[0] - 0.5, endPt[1] - 0.5),
                lineWidth,
                method=method,
            )
            return profile

        currentData1 = item.getImageData1()
        currentData2 = item.getImageData2()
        profile1 = computeProfile(currentData1)
        profile2 = computeProfile(currentData2)

        # Compute the line size
        lineSize = numpy.sqrt(
            (roiEnd[1] - roiStart[1]) ** 2 + (roiEnd[0] - roiStart[0]) ** 2
        )
        coords = numpy.linspace(
            0, lineSize, len(profile1), endpoint=True, dtype=numpy.float32
        )

        title = rois._lineProfileTitle(*roiStart, *roiEnd)
        title = title + "; width = %d" % lineWidth
        xLabel = "√({xlabel}²+{ylabel}²)"
        yLabel = str(method).capitalize()

        # Use the axis names from the original plot
        profileManager = self.getProfileManager()
        plot = profileManager.getPlotWidget()
        xLabel = rois._relabelAxes(plot, xLabel)
        title = rois._relabelAxes(plot, title)

        data = core.CurvesProfileData(
            coords=coords,
            profiles=[
                core.CurveProfileDesc(profile1, color=COLOR_A, name="profileA"),
                core.CurveProfileDesc(profile2, color=COLOR_B, name="profileB"),
            ],
            title=title,
            xLabel=xLabel,
            yLabel=yLabel,
        )
        return data



# /*##########################################################################
#
# Copyright (c) 2014-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
This modules provides the rendering of plot titles, axes and grid.
"""

from __future__ import annotations

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "03/04/2017"


# TODO
# keep aspect ratio managed here?
# smarter dirty flag handling?

import datetime as dt
import math
import weakref
import logging
import numbers
from typing import Optional, Union
from collections import namedtuple

import numpy

from .... import qt
from ...._glutils import gl, Program
from ....utils.matplotlib import DefaultTickFormatter
from ..._utils import checkAxisLimits, FLOAT32_MINPOS
from .GLSupport import mat4Ortho
from .GLText import Text2D, CENTER, BOTTOM, TOP, LEFT, RIGHT, ROTATE_270
from ..._utils.ticklayout import niceNumbersAdaptative, niceNumbersForLog10
from ..._utils.dtime_ticklayout import (
    DtUnit,
    bestUnit,
    calcTicksAdaptive,
    formatDatetimes,
)
from ..._utils.dtime_ticklayout import timestamp

_logger = logging.getLogger(__name__)


# PlotAxis ####################################################################


class PlotAxis(object):
    """Represents a 1D axis of the plot.
    This class is intended to be used with :class:`GLPlotFrame`.
    """

    def __init__(
        self,
        plotFrame,
        tickLength=(0.0, 0.0),
        foregroundColor=(0.0, 0.0, 0.0, 1.0),
        labelAlign=CENTER,
        labelVAlign=CENTER,
        titleAlign=CENTER,
        titleVAlign=CENTER,
        orderOffsetAlign=CENTER,
        orderOffsetVAlign=CENTER,
        titleRotate=0,
        titleOffset=(0.0, 0.0),
        font: qt.QFont | None = None,
    ):
        self._tickFormatter = DefaultTickFormatter()
        self._ticks = None
        self._orderAndOffsetText = ""

        self._plotFrameRef = weakref.ref(plotFrame)

        self._isDateTime = False
        self._timeZone = None
        self._isLog = False
        self._dataRange = 1.0, 100.0
        self._displayCoords = (0.0, 0.0), (1.0, 0.0)
        self._title = ""

        self._tickLength = tickLength
        self._foregroundColor = foregroundColor
        self._labelAlign = labelAlign
        self._labelVAlign = labelVAlign
        self._orderOffetAnchor = (1.0, 0.0)
        self._orderOffsetAlign = orderOffsetAlign
        self._orderOffsetVAlign = orderOffsetVAlign
        self._titleAlign = titleAlign
        self._titleVAlign = titleVAlign
        self._titleRotate = titleRotate
        self._titleOffset = titleOffset
        self._font = font

    @property
    def dataRange(self):
        """The range of the data represented on the axis as a tuple
        of 2 floats: (min, max)."""
        return self._dataRange

    @property
    def font(self) -> qt.QFont:
        if self._font is None:
            return qt.QApplication.instance().font()
        return self._font

    @dataRange.setter
    def dataRange(self, dataRange):
        assert len(dataRange) == 2
        assert dataRange[0] <= dataRange[1]
        dataRange = float(dataRange[0]), float(dataRange[1])

        if dataRange != self._dataRange:
            self._dataRange = dataRange
            self._dirtyTicks()

    @property
    def isLog(self):
        """Whether the axis is using a log10 scale or not as a bool."""
        return self._isLog

    @isLog.setter
    def isLog(self, isLog):
        isLog = bool(isLog)
        if isLog != self._isLog:
            self._isLog = isLog
            self._dirtyTicks()

    @property
    def timeZone(self):
        """Returnss datetime.tzinfo that is used if this axis plots date times."""
        return self._timeZone

    @timeZone.setter
    def timeZone(self, tz):
        """Sets dateetime.tzinfo that is used if this axis plots date times."""
        self._timeZone = tz
        self._dirtyTicks()

    @property
    def isTimeSeries(self):
        """Whether the axis is showing floats as datetime objects"""
        return self._isDateTime

    @isTimeSeries.setter
    def isTimeSeries(self, isTimeSeries):
        isTimeSeries = bool(isTimeSeries)
        if isTimeSeries != self._isDateTime:
            self._isDateTime = isTimeSeries
            self._dirtyTicks()

    @property
    def displayCoords(self):
        """The coordinates of the start and end points of the axis
        in display space (i.e., in pixels) as a tuple of 2 tuples of
        2 floats: ((x0, y0), (x1, y1)).
        """
        return self._displayCoords

    @displayCoords.setter
    def displayCoords(self, displayCoords):
        assert len(displayCoords) == 2
        assert len(displayCoords[0]) == 2
        assert len(displayCoords[1]) == 2
        displayCoords = tuple(displayCoords[0]), tuple(displayCoords[1])
        if displayCoords != self._displayCoords:
            self._displayCoords = displayCoords
            self._dirtyTicks()

    @property
    def devicePixelRatio(self):
        """Returns the ratio between qt pixels and device pixels."""
        plotFrame = self._plotFrameRef()
        return plotFrame.devicePixelRatio if plotFrame is not None else 1.0

    @property
    def dotsPerInch(self):
        """Returns the screen DPI"""
        plotFrame = self._plotFrameRef()
        return plotFrame.dotsPerInch if plotFrame is not None else 92

    @property
    def title(self):
        """The text label associated with this axis as a str in latin-1."""
        return self._title

    @title.setter
    def title(self, title):
        if title != self._title:
            self._title = title
            self._dirtyPlotFrame()

    @property
    def orderOffetAnchor(self) -> tuple[float, float]:
        """Anchor position for the tick order&offset text"""
        return self._orderOffetAnchor

    @orderOffetAnchor.setter
    def orderOffetAnchor(self, position: tuple[float, float]):
        if position != self._orderOffetAnchor:
            self._orderOffetAnchor = position
            self._dirtyTicks()

    @property
    def titleOffset(self):
        """Title offset in pixels (x: int, y: int)"""
        return self._titleOffset

    @titleOffset.setter
    def titleOffset(self, offset):
        if offset != self._titleOffset:
            self._titleOffset = offset
            self._dirtyTicks()

    @property
    def foregroundColor(self):
        """Color used for frame and labels"""
        return self._foregroundColor

    @foregroundColor.setter
    def foregroundColor(self, color):
        """Color used for frame and labels"""
        assert len(color) == 4, "foregroundColor must have length 4, got {}".format(
            len(self._foregroundColor)
        )
        if self._foregroundColor != color:
            self._foregroundColor = color
            self._dirtyTicks()

    @property
    def ticks(self):
        """Ticks as tuples: ((x, y) in display, dataPos, textLabel)."""
        if self._ticks is None:
            self._ticks = tuple(self._ticksGenerator())
        return self._ticks

    def getVerticesAndLabels(self):
        """Create the list of vertices for axis and associated text labels.

        :returns: A tuple: List of 2D line vertices, List of Text2D labels.
        """
        vertices = list(self.displayCoords)  # Add start and end points
        labels = []

        xTickLength, yTickLength = self._tickLength
        xTickLength *= self.devicePixelRatio
        yTickLength *= self.devicePixelRatio
        for (xPixel, yPixel), dataPos, text in self.ticks:
            if text is None:
                tickScale = 0.5
            else:
                tickScale = 1.0

                label = Text2D(
                    text=text,
                    font=self.font,
                    color=self._foregroundColor,
                    x=xPixel - xTickLength,
                    y=yPixel - yTickLength,
                    align=self._labelAlign,
                    valign=self._labelVAlign,
                    devicePixelRatio=self.devicePixelRatio,
                )
                labels.append(label)

            vertices.append((xPixel, yPixel))
            vertices.append(
                (xPixel + tickScale * xTickLength, yPixel + tickScale * yTickLength)
            )

        (x0, y0), (x1, y1) = self.displayCoords
        xAxisCenter = 0.5 * (x0 + x1)
        yAxisCenter = 0.5 * (y0 + y1)

        xOffset, yOffset = self.titleOffset

        # Adaptative title positioning:
        # tickNorm = math.sqrt(xTickLength ** 2 + yTickLength ** 2)
        # xOffset = -tickLabelsSize[0] * xTickLength / tickNorm
        # xOffset -= 3 * xTickLength
        # yOffset = -tickLabelsSize[1] * yTickLength / tickNorm
        # yOffset -= 3 * yTickLength

        axisTitle = Text2D(
            text=self.title,
            font=self.font,
            color=self._foregroundColor,
            x=xAxisCenter + xOffset,
            y=yAxisCenter + yOffset,
            align=self._titleAlign,
            valign=self._titleVAlign,
            rotate=self._titleRotate,
            devicePixelRatio=self.devicePixelRatio,
        )
        labels.append(axisTitle)

        if self._orderAndOffsetText:
            xOrderOffset, yOrderOffet = self.orderOffetAnchor
            labels.append(
                Text2D(
                    text=self._orderAndOffsetText,
                    font=self.font,
                    color=self._foregroundColor,
                    x=xOrderOffset,
                    y=yOrderOffet,
                    align=self._orderOffsetAlign,
                    valign=self._orderOffsetVAlign,
                    devicePixelRatio=self.devicePixelRatio,
                )
            )
        return vertices, labels

    def _dirtyPlotFrame(self):
        """Dirty parent GLPlotFrame"""
        plotFrame = self._plotFrameRef()
        if plotFrame is not None:
            plotFrame._dirty()

    def _dirtyTicks(self):
        """Mark ticks as dirty and notify listener (i.e., background)."""
        self._ticks = None
        self._dirtyPlotFrame()

    @staticmethod
    def _frange(start, stop, step):
        """range for float (including stop)."""
        while start <= stop:
            yield start
            start += step

    def _ticksGenerator(self):
        """Generator of ticks as tuples:
        ((x, y) in display, dataPos, textLabel).
        """
        self._orderAndOffsetText = ""

        dataMin, dataMax = self.dataRange
        if self.isLog and dataMin <= 0.0:
            _logger.warning("Getting ticks while isLog=True and dataRange[0]<=0.")
            dataMin = 1.0
            if dataMax < dataMin:
                dataMax = 1.0

        if dataMin != dataMax:  # data range is not null
            (x0, y0), (x1, y1) = self.displayCoords

            if self.isLog:
                if self.isTimeSeries:
                    _logger.warning("Time series not implemented for log-scale")

                logMin, logMax = math.log10(dataMin), math.log10(dataMax)
                tickMin, tickMax, step, _ = niceNumbersForLog10(logMin, logMax)

                xScale = (x1 - x0) / (logMax - logMin)
                yScale = (y1 - y0) / (logMax - logMin)

                for logPos in self._frange(tickMin, tickMax, step):
                    if logMin <= logPos <= logMax:
                        dataPos = 10**logPos
                        xPixel = x0 + (logPos - logMin) * xScale
                        yPixel = y0 + (logPos - logMin) * yScale
                        text = "1e%+03d" % logPos
                        yield ((xPixel, yPixel), dataPos, text)

                if step == 1:
                    ticks = list(self._frange(tickMin, tickMax, step))[:-1]
                    for logPos in ticks:
                        dataOrigPos = 10**logPos
                        for index in range(2, 10):
                            dataPos = dataOrigPos * index
                            if dataMin <= dataPos <= dataMax:
                                logSubPos = math.log10(dataPos)
                                xPixel = x0 + (logSubPos - logMin) * xScale
                                yPixel = y0 + (logSubPos - logMin) * yScale
                                yield ((xPixel, yPixel), dataPos, None)

            else:
                xScale = (x1 - x0) / (dataMax - dataMin)
                yScale = (y1 - y0) / (dataMax - dataMin)

                nbPixels = (
                    math.sqrt(pow(x1 - x0, 2) + pow(y1 - y0, 2)) / self.devicePixelRatio
                )

                # Density of 1.3 label per 92 pixels
                # i.e., 1.3 label per inch on a 92 dpi screen
                tickDensity = 1.3 * self.devicePixelRatio / self.dotsPerInch

                if not self.isTimeSeries:
                    tickMin, tickMax, step, _ = niceNumbersAdaptative(
                        dataMin, dataMax, nbPixels, tickDensity
                    )

                    visibleTickPositions = [
                        pos
                        for pos in self._frange(tickMin, tickMax, step)
                        if dataMin <= pos <= dataMax
                    ]
                    self._tickFormatter.axis.set_view_interval(dataMin, dataMax)
                    self._tickFormatter.axis.set_data_interval(dataMin, dataMax)
                    texts = self._tickFormatter.format_ticks(visibleTickPositions)
                    self._orderAndOffsetText = self._tickFormatter.get_offset()

                    for dataPos, text in zip(visibleTickPositions, texts):
                        xPixel = x0 + (dataPos - dataMin) * xScale
                        yPixel = y0 + (dataPos - dataMin) * yScale
                        yield ((xPixel, yPixel), dataPos, text)

                else:
                    # Time series
                    try:
                        dtMin = dt.datetime.fromtimestamp(dataMin, tz=self.timeZone)
                        dtMax = dt.datetime.fromtimestamp(dataMax, tz=self.timeZone)
                    except ValueError:
                        _logger.warning("Data range cannot be displayed with time axis")
                        return  # Range is out of bound of the datetime

                    if bestUnit(
                        (dtMax - dtMin).total_seconds() == DtUnit.MICRO_SECONDS
                    ):
                        # Special case for micro seconds: Reduce tick density
                        tickDensity = 1.0 * self.devicePixelRatio / self.dotsPerInch

                    tickDateTimes, spacing, unit = calcTicksAdaptive(
                        dtMin, dtMax, nbPixels, tickDensity
                    )
                    visibleDatetimes = tuple(
                        dt for dt in tickDateTimes if dtMin <= dt <= dtMax
                    )
                    ticks = formatDatetimes(visibleDatetimes, spacing, unit)

                    for tickDateTime, text in ticks.items():
                        dataPos = timestamp(tickDateTime)
                        xPixel = x0 + (dataPos - dataMin) * xScale
                        yPixel = y0 + (dataPos - dataMin) * yScale
                        yield ((xPixel, yPixel), dataPos, text)


# GLPlotFrame #################################################################


class GLPlotFrame(object):
    """Base class for rendering a 2D frame surrounded by axes."""

    _TICK_LENGTH_IN_PIXELS = 5
    _LINE_WIDTH = 1

    _SHADERS = {
        "vertex": """
    attribute vec2 position;
    uniform mat4 matrix;

    void main(void) {
        gl_Position = matrix * vec4(position, 0.0, 1.0);
    }
    """,
        "fragment": """
    uniform vec4 color;
    uniform float tickFactor; /* = 1./tickLength or 0. for solid line */

    void main(void) {
        if (mod(tickFactor * (gl_FragCoord.x + gl_FragCoord.y), 2.) < 1.) {
            gl_FragColor = color;
        } else {
            discard;
        }
    }
    """,
    }

    _Margins = namedtuple("Margins", ("left", "right", "top", "bottom"))

    # Margins used when plot frame is not displayed
    _NoDisplayMargins = _Margins(0, 0, 0, 0)

    def __init__(self, marginRatios, foregroundColor, gridColor, font: qt.QFont):
        """
        :param List[float] marginRatios:
            The ratios of margins around plot area for axis and labels.
            (left, top, right, bottom) as float in [0., 1.]
        :param foregroundColor: color used for the frame and labels.
        :type foregroundColor: tuple with RGBA values ranging from 0.0 to 1.0
        :param gridColor: color used for grid lines.
        :type gridColor: tuple RGBA with RGBA values ranging from 0.0 to 1.0
        :param font: Font used by the axes label
        """
        self._renderResources = None

        self.__marginRatios = marginRatios
        self.__marginsCache = None

        self._foregroundColor = foregroundColor
        self._gridColor = gridColor

        self.axes = []  # List of PlotAxis to be updated by subclasses

        self._grid = False
        self._size = 0.0, 0.0
        self._title = ""
        self._font: qt.QFont = font

        self._devicePixelRatio = 1.0
        self._dpi = 92

    @property
    def isDirty(self):
        """True if it need to refresh graphic rendering, False otherwise."""
        return self._renderResources is None

    GRID_NONE = 0
    GRID_MAIN_TICKS = 1
    GRID_SUB_TICKS = 2
    GRID_ALL_TICKS = GRID_MAIN_TICKS + GRID_SUB_TICKS

    @property
    def foregroundColor(self):
        """Color used for frame and labels"""
        return self._foregroundColor

    @foregroundColor.setter
    def foregroundColor(self, color):
        """Color used for frame and labels"""
        assert len(color) == 4, "foregroundColor must have length 4, got {}".format(
            len(self._foregroundColor)
        )
        if self._foregroundColor != color:
            self._foregroundColor = color
            for axis in self.axes:
                axis.foregroundColor = color
            self._dirty()

    @property
    def gridColor(self):
        """Color used for frame and labels"""
        return self._gridColor

    @gridColor.setter
    def gridColor(self, color):
        """Color used for frame and labels"""
        assert len(color) == 4, "gridColor must have length 4, got {}".format(
            len(self._gridColor)
        )
        if self._gridColor != color:
            self._gridColor = color
            self._dirty()

    @property
    def marginRatios(self):
        """Plot margin ratios: (left, top, right, bottom) as 4 float in [0, 1]."""
        return self.__marginRatios

    @marginRatios.setter
    def marginRatios(self, ratios):
        ratios = tuple(float(v) for v in ratios)
        assert len(ratios) == 4
        for value in ratios:
            assert 0.0 <= value <= 1.0
        assert ratios[0] + ratios[2] < 1.0
        assert ratios[1] + ratios[3] < 1.0

        if self.__marginRatios != ratios:
            self.__marginRatios = ratios
            self.__marginsCache = None  # Clear cached margins
            self._dirty()

    @property
    def margins(self):
        """Margins in pixels around the plot."""
        if self.__marginsCache is None:
            width, height = self.size
            left, top, right, bottom = self.marginRatios
            self.__marginsCache = self._Margins(
                left=int(left * width),
                right=int(right * width),
                top=int(top * height),
                bottom=int(bottom * height),
            )
        return self.__marginsCache

    @property
    def devicePixelRatio(self):
        return self._devicePixelRatio

    @devicePixelRatio.setter
    def devicePixelRatio(self, ratio):
        if ratio != self._devicePixelRatio:
            self._devicePixelRatio = ratio
            self._dirty()

    @property
    def dotsPerInch(self):
        return self._dpi

    @dotsPerInch.setter
    def dotsPerInch(self, dpi):
        if dpi != self._dpi:
            self._dpi = dpi
            self._dirty()

    @property
    def grid(self):
        """Grid display mode:
        - 0: No grid.
        - 1: Grid on main ticks.
        - 2: Grid on sub-ticks for log scale axes.
        - 3: Grid on main and sub ticks."""
        return self._grid

    @grid.setter
    def grid(self, grid):
        assert grid in (
            self.GRID_NONE,
            self.GRID_MAIN_TICKS,
            self.GRID_SUB_TICKS,
            self.GRID_ALL_TICKS,
        )
        if grid != self._grid:
            self._grid = grid
            self._dirty()

    @property
    def size(self):
        """Size in device pixels of the plot area including margins."""
        return self._size

    @size.setter
    def size(self, size):
        assert len(size) == 2
        size = tuple(size)
        if size != self._size:
            self._size = size
            self.__marginsCache = None  # Clear cached margins
            self._dirty()

    @property
    def plotOrigin(self):
        """Plot area origin (left, top) in widget coordinates in pixels."""
        return self.margins.left, self.margins.top

    @property
    def plotSize(self):
        """Plot area size (width, height) in pixels."""
        w, h = self.size
        w -= self.margins.left + self.margins.right
        h -= self.margins.top + self.margins.bottom
        return w, h

    @property
    def title(self):
        """Main title as a str in latin-1."""
        return self._title

    @title.setter
    def title(self, title):
        if title != self._title:
            self._title = title
            self._dirty()

        # In-place update
        # if self._renderResources is not None:
        #    self._renderResources[-1][-1].text = title

    def _dirty(self):
        # When Text2D require discard we need to handle it
        self._renderResources = None

    def _buildGridVertices(self):
        if self._grid == self.GRID_NONE:
            return []

        elif self._grid == self.GRID_MAIN_TICKS:

            def test(text):
                return text is not None

        elif self._grid == self.GRID_SUB_TICKS:

            def test(text):
                return text is None

        elif self._grid == self.GRID_ALL_TICKS:

            def test(_):
                return True

        else:
            logging.warning("Wrong grid mode: %d" % self._grid)
            return []

        return self._buildGridVerticesWithTest(test)

    def _buildGridVerticesWithTest(self, test):
        """Override in subclass to generate grid vertices"""
        return []

    def _buildVerticesAndLabels(self):
        # To fill with copy of axes lists
        vertices = []
        labels = []

        for axis in self.axes:
            axisVertices, axisLabels = axis.getVerticesAndLabels()
            vertices += axisVertices
            labels += axisLabels

        vertices = numpy.array(vertices, dtype=numpy.float32)

        # Add main title
        xTitle = (self.size[0] + self.margins.left - self.margins.right) // 2
        yTitle = self.margins.top - self._TICK_LENGTH_IN_PIXELS
        labels.append(
            Text2D(
                text=self.title,
                font=self._font,
                color=self._foregroundColor,
                x=xTitle,
                y=yTitle,
                align=CENTER,
                valign=BOTTOM,
                devicePixelRatio=self.devicePixelRatio,
            )
        )

        # grid
        gridVertices = numpy.array(self._buildGridVertices(), dtype=numpy.float32)

        self._renderResources = (vertices, gridVertices, labels)

    _program = Program(_SHADERS["vertex"], _SHADERS["fragment"], attrib0="position")

    def render(self):
        if self.margins == self._NoDisplayMargins:
            return

        if self._renderResources is None:
            self._buildVerticesAndLabels()
        vertices, gridVertices, labels = self._renderResources

        width, height = self.size
        matProj = mat4Ortho(0, width, height, 0, 1, -1)

        gl.glViewport(0, 0, width, height)

        prog = self._program
        prog.use()

        gl.glLineWidth(self._LINE_WIDTH)

        gl.glUniformMatrix4fv(
            prog.uniforms["matrix"], 1, gl.GL_TRUE, matProj.astype(numpy.float32)
        )
        gl.glUniform4f(prog.uniforms["color"], *self._foregroundColor)
        gl.glUniform1f(prog.uniforms["tickFactor"], 0.0)

        gl.glEnableVertexAttribArray(prog.attributes["position"])
        gl.glVertexAttribPointer(
            prog.attributes["position"], 2, gl.GL_FLOAT, gl.GL_FALSE, 0, vertices
        )

        gl.glDrawArrays(gl.GL_LINES, 0, len(vertices))

        for label in labels:
            label.render(matProj, self.dotsPerInch)

    def renderGrid(self):
        if self._grid == self.GRID_NONE:
            return

        if self._renderResources is None:
            self._buildVerticesAndLabels()
        vertices, gridVertices, labels = self._renderResources

        width, height = self.size
        matProj = mat4Ortho(0, width, height, 0, 1, -1)

        gl.glViewport(0, 0, width, height)

        prog = self._program
        prog.use()

        gl.glLineWidth(self._LINE_WIDTH)
        gl.glUniformMatrix4fv(
            prog.uniforms["matrix"], 1, gl.GL_TRUE, matProj.astype(numpy.float32)
        )
        gl.glUniform4f(prog.uniforms["color"], *self._gridColor)
        gl.glUniform1f(prog.uniforms["tickFactor"], 0.0)  # 1/2.)  # 1/tickLen

        gl.glEnableVertexAttribArray(prog.attributes["position"])
        gl.glVertexAttribPointer(
            prog.attributes["position"], 2, gl.GL_FLOAT, gl.GL_FALSE, 0, gridVertices
        )

        gl.glDrawArrays(gl.GL_LINES, 0, len(gridVertices))


# GLPlotFrame2D ###############################################################


class GLPlotFrame2D(GLPlotFrame):
    def __init__(self, marginRatios, foregroundColor, gridColor, font: qt.QFont):
        """
        :param List[float] marginRatios:
            The ratios of margins around plot area for axis and labels.
            (left, top, right, bottom) as float in [0., 1.]
        :param foregroundColor: color used for the frame and labels.
        :type foregroundColor: tuple with RGBA values ranging from 0.0 to 1.0
        :param gridColor: color used for grid lines.
        :type gridColor: tuple RGBA with RGBA values ranging from 0.0 to 1.0
        :param font: Font used by the axes label
        """
        super(GLPlotFrame2D, self).__init__(
            marginRatios, foregroundColor, gridColor, font
        )
        self._font = font

        self.axes.append(
            PlotAxis(
                self,
                tickLength=(0.0, -5.0),
                foregroundColor=self._foregroundColor,
                labelAlign=CENTER,
                labelVAlign=TOP,
                orderOffsetAlign=RIGHT,
                orderOffsetVAlign=TOP,
                titleAlign=CENTER,
                titleVAlign=TOP,
                titleRotate=0,
                font=self._font,
            )
        )

        self._x2AxisCoords = ()

        self.axes.append(
            PlotAxis(
                self,
                tickLength=(5.0, 0.0),
                foregroundColor=self._foregroundColor,
                labelAlign=RIGHT,
                labelVAlign=CENTER,
                orderOffsetAlign=LEFT,
                orderOffsetVAlign=BOTTOM,
                titleAlign=CENTER,
                titleVAlign=BOTTOM,
                titleRotate=ROTATE_270,
                font=self._font,
            )
        )

        self._y2Axis = PlotAxis(
            self,
            tickLength=(-5.0, 0.0),
            foregroundColor=self._foregroundColor,
            labelAlign=LEFT,
            labelVAlign=CENTER,
            orderOffsetAlign=RIGHT,
            orderOffsetVAlign=BOTTOM,
            titleAlign=CENTER,
            titleVAlign=TOP,
            titleRotate=ROTATE_270,
            font=self._font,
        )

        self._isYAxisInverted = False

        self._dataRanges = {"x": (1.0, 100.0), "y": (1.0, 100.0), "y2": (1.0, 100.0)}

        self._baseVectors = (1.0, 0.0), (0.0, 1.0)

        self._transformedDataRanges = None
        self._transformedDataProjMat = None
        self._transformedDataY2ProjMat = None

    def _dirty(self):
        super(GLPlotFrame2D, self)._dirty()
        self._transformedDataRanges = None
        self._transformedDataProjMat = None
        self._transformedDataY2ProjMat = None

    @property
    def isDirty(self):
        """True if it need to refresh graphic rendering, False otherwise."""
        return (
            super(GLPlotFrame2D, self).isDirty
            or self._transformedDataRanges is None
            or self._transformedDataProjMat is None
            or self._transformedDataY2ProjMat is None
        )

    @property
    def xAxis(self):
        return self.axes[0]

    @property
    def yAxis(self):
        return self.axes[1]

    @property
    def y2Axis(self):
        return self._y2Axis

    @property
    def isY2Axis(self):
        """Whether to display the left Y axis or not."""
        return len(self.axes) == 3

    @isY2Axis.setter
    def isY2Axis(self, isY2Axis):
        if isY2Axis != self.isY2Axis:
            if isY2Axis:
                self.axes.append(self._y2Axis)
            else:
                self.axes = self.axes[:2]

            self._dirty()

    @property
    def isYAxisInverted(self):
        """Whether Y axes are inverted or not as a bool."""
        return self._isYAxisInverted

    @isYAxisInverted.setter
    def isYAxisInverted(self, value):
        value = bool(value)
        if value != self._isYAxisInverted:
            self._isYAxisInverted = value
            self._dirty()

    DEFAULT_BASE_VECTORS = (1.0, 0.0), (0.0, 1.0)
    """Values of baseVectors for orthogonal axes."""

    @property
    def baseVectors(self):
        """Coordinates of the X and Y axes in the orthogonal plot coords.

        Raises ValueError if corresponding matrix is singular.

        2 tuples of 2 floats: (xx, xy), (yx, yy)
        """
        return self._baseVectors

    @baseVectors.setter
    def baseVectors(self, baseVectors):
        self._dirty()

        (xx, xy), (yx, yy) = baseVectors
        vectors = (float(xx), float(xy)), (float(yx), float(yy))

        det = vectors[0][0] * vectors[1][1] - vectors[1][0] * vectors[0][1]
        if det == 0.0:
            raise ValueError("Singular matrix for base vectors: " + str(vectors))

        if vectors != self._baseVectors:
            self._baseVectors = vectors
            self._dirty()

    def _updateTitleOffset(self):
        """Update axes title offset according to margins"""
        margins = self.margins
        self.xAxis.titleOffset = 0, margins.bottom // 2
        self.yAxis.titleOffset = -3 * margins.left // 4, 0
        self.y2Axis.titleOffset = 3 * margins.right // 4, 0

    # Override size and marginRatios setters to update titleOffsets
    @GLPlotFrame.size.setter
    def size(self, size):
        GLPlotFrame.size.fset(self, size)
        self._updateTitleOffset()

    @GLPlotFrame.marginRatios.setter
    def marginRatios(self, ratios):
        GLPlotFrame.marginRatios.fset(self, ratios)
        self._updateTitleOffset()

    @property
    def dataRanges(self):
        """Ranges of data visible in the plot on x, y and y2 axes.

        This is different to the axes range when axes are not orthogonal.

        Type: ((xMin, xMax), (yMin, yMax), (y2Min, y2Max))
        """
        return self._DataRanges(
            self._dataRanges["x"], self._dataRanges["y"], self._dataRanges["y2"]
        )

    def setDataRanges(self, x=None, y=None, y2=None):
        """Set data range over each axes.

        The provided ranges are clipped to possible values
        (i.e., 32 float range + positive range for log scale).

        :param x: (min, max) data range over X axis
        :param y: (min, max) data range over Y axis
        :param y2: (min, max) data range over Y2 axis
        """
        if x is not None:
            self._dataRanges["x"] = checkAxisLimits(
                x[0], x[1], self.xAxis.isLog, name="x"
            )

        if y is not None:
            self._dataRanges["y"] = checkAxisLimits(
                y[0], y[1], self.yAxis.isLog, name="y"
            )

        if y2 is not None:
            self._dataRanges["y2"] = checkAxisLimits(
                y2[0], y2[1], self.y2Axis.isLog, name="y2"
            )

        self.xAxis.dataRange = self._dataRanges["x"]
        self.yAxis.dataRange = self._dataRanges["y"]
        self.y2Axis.dataRange = self._dataRanges["y2"]

    _DataRanges = namedtuple("dataRanges", ("x", "y", "y2"))

    @property
    def transformedDataRanges(self):
        """Bounds of the displayed area in transformed data coordinates
        (i.e., log scale applied if any as well as skew)

        3-tuple of 2-tuple (min, max) for each axis: x, y, y2.
        """
        if self._transformedDataRanges is None:
            (xMin, xMax), (yMin, yMax), (y2Min, y2Max) = self.dataRanges

            if self.xAxis.isLog:
                try:
                    xMin = math.log10(xMin)
                except ValueError:
                    _logger.info("xMin: warning log10(%f)", xMin)
                    xMin = 0.0
                try:
                    xMax = math.log10(xMax)
                except ValueError:
                    _logger.info("xMax: warning log10(%f)", xMax)
                    xMax = 0.0

            if self.yAxis.isLog:
                try:
                    yMin = math.log10(yMin)
                except ValueError:
                    _logger.info("yMin: warning log10(%f)", yMin)
                    yMin = 0.0
                try:
                    yMax = math.log10(yMax)
                except ValueError:
                    _logger.info("yMax: warning log10(%f)", yMax)
                    yMax = 0.0

                try:
                    y2Min = math.log10(y2Min)
                except ValueError:
                    _logger.info("yMin: warning log10(%f)", y2Min)
                    y2Min = 0.0
                try:
                    y2Max = math.log10(y2Max)
                except ValueError:
                    _logger.info("yMax: warning log10(%f)", y2Max)
                    y2Max = 0.0

            self._transformedDataRanges = self._DataRanges(
                (xMin, xMax), (yMin, yMax), (y2Min, y2Max)
            )

        return self._transformedDataRanges

    @property
    def transformedDataProjMat(self):
        """Orthographic projection matrix for rendering transformed data

        :type: numpy.matrix
        """
        if self._transformedDataProjMat is None:
            xMin, xMax = self.transformedDataRanges.x
            yMin, yMax = self.transformedDataRanges.y

            if self.isYAxisInverted:
                mat = mat4Ortho(xMin, xMax, yMax, yMin, 1, -1)
            else:
                mat = mat4Ortho(xMin, xMax, yMin, yMax, 1, -1)
            self._transformedDataProjMat = mat

        return self._transformedDataProjMat

    @property
    def transformedDataY2ProjMat(self):
        """Orthographic projection matrix for rendering transformed data
        for the 2nd Y axis

        :type: numpy.matrix
        """
        if self._transformedDataY2ProjMat is None:
            xMin, xMax = self.transformedDataRanges.x
            y2Min, y2Max = self.transformedDataRanges.y2

            if self.isYAxisInverted:
                mat = mat4Ortho(xMin, xMax, y2Max, y2Min, 1, -1)
            else:
                mat = mat4Ortho(xMin, xMax, y2Min, y2Max, 1, -1)
            self._transformedDataY2ProjMat = mat

        return self._transformedDataY2ProjMat

    @staticmethod
    def __applyLog(
        data: Union[float, numpy.ndarray], isLog: bool
    ) -> Optional[Union[float, numpy.ndarray]]:
        """Apply log to data filtering out"""
        if not isLog:
            return data

        if isinstance(data, numbers.Real):
            return None if data < FLOAT32_MINPOS else math.log10(data)

        isBelowMin = data < FLOAT32_MINPOS
        if numpy.any(isBelowMin):
            data = numpy.array(data, copy=True, dtype=numpy.float64)
            data[isBelowMin] = numpy.nan

        with numpy.errstate(divide="ignore"):
            return numpy.log10(data)

    def dataToPixel(self, x, y, axis="left"):
        """Convert data coordinate to widget pixel coordinate."""
        assert axis in ("left", "right")

        trBounds = self.transformedDataRanges

        xDataTr = self.__applyLog(x, self.xAxis.isLog)
        if xDataTr is None:
            return None

        yDataTr = self.__applyLog(y, self.yAxis.isLog)
        if yDataTr is None:
            return None

        # Non-orthogonal axes
        if self.baseVectors != self.DEFAULT_BASE_VECTORS:
            (xx, xy), (yx, yy) = self.baseVectors
            skew_mat = numpy.array(((xx, yx), (xy, yy)))

            coords = numpy.dot(skew_mat, numpy.array((xDataTr, yDataTr)))
            xDataTr, yDataTr = coords

        plotWidth, plotHeight = self.plotSize

        xPixel = self.margins.left + plotWidth * (xDataTr - trBounds.x[0]) / (
            trBounds.x[1] - trBounds.x[0]
        )

        usedAxis = trBounds.y if axis == "left" else trBounds.y2
        yOffset = plotHeight * (yDataTr - usedAxis[0]) / (usedAxis[1] - usedAxis[0])

        if self.isYAxisInverted:
            yPixel = self.margins.top + yOffset
        else:
            yPixel = self.size[1] - self.margins.bottom - yOffset

        return (
            int(xPixel)
            if isinstance(xPixel, numbers.Real)
            else xPixel.astype(numpy.int64),
            int(yPixel)
            if isinstance(yPixel, numbers.Real)
            else yPixel.astype(numpy.int64),
        )

    def pixelToData(self, x, y, axis="left"):
        """Convert pixel position to data coordinates.

        :param float x: X coord
        :param float y: Y coord
        :param str axis: Y axis to use in ('left', 'right')
        :return: (x, y) position in data coords
        """
        assert axis in ("left", "right")

        plotWidth, plotHeight = self.plotSize

        trBounds = self.transformedDataRanges

        xData = (x - self.margins.left + 0.5) / float(plotWidth)
        xData = trBounds.x[0] + xData * (trBounds.x[1] - trBounds.x[0])

        usedAxis = trBounds.y if axis == "left" else trBounds.y2
        if self.isYAxisInverted:
            yData = (y - self.margins.top + 0.5) / float(plotHeight)
            yData = usedAxis[0] + yData * (usedAxis[1] - usedAxis[0])
        else:
            yData = self.size[1] - self.margins.bottom - y - 0.5
            yData /= float(plotHeight)
            yData = usedAxis[0] + yData * (usedAxis[1] - usedAxis[0])

        # non-orthogonal axis
        if self.baseVectors != self.DEFAULT_BASE_VECTORS:
            (xx, xy), (yx, yy) = self.baseVectors
            skew_mat = numpy.array(((xx, yx), (xy, yy)))
            skew_mat = numpy.linalg.inv(skew_mat)

            coords = numpy.dot(skew_mat, numpy.array((xData, yData)))
            xData, yData = coords

        if self.xAxis.isLog:
            xData = pow(10, xData)
        if self.yAxis.isLog:
            yData = pow(10, yData)

        return xData, yData

    def _buildGridVerticesWithTest(self, test):
        vertices = []

        if self.baseVectors == self.DEFAULT_BASE_VECTORS:
            for axis in self.axes:
                for (xPixel, yPixel), data, text in axis.ticks:
                    if test(text):
                        vertices.append((xPixel, yPixel))
                        if axis == self.xAxis:
                            vertices.append((xPixel, self.margins.top))
                        elif axis == self.yAxis:
                            vertices.append((self.size[0] - self.margins.right, yPixel))
                        else:  # axis == self.y2Axis
                            vertices.append((self.margins.left, yPixel))

        else:
            # Get plot corners in data coords
            plotLeft, plotTop = self.plotOrigin
            plotWidth, plotHeight = self.plotSize

            corners = [
                (plotLeft, plotTop),
                (plotLeft, plotTop + plotHeight),
                (plotLeft + plotWidth, plotTop + plotHeight),
                (plotLeft + plotWidth, plotTop),
            ]

            for axis in self.axes:
                if axis == self.xAxis:
                    cornersInData = numpy.array(
                        [self.pixelToData(x, y) for (x, y) in corners]
                    )
                    borders = (
                        (cornersInData[0], cornersInData[3]),  # top
                        (cornersInData[1], cornersInData[0]),  # left
                        (cornersInData[3], cornersInData[2]),
                    )  # right

                    for (xPixel, yPixel), data, text in axis.ticks:
                        if test(text):
                            for (x0, y0), (x1, y1) in borders:
                                if min(x0, x1) <= data < max(x0, x1):
                                    yIntersect = (data - x0) * (y1 - y0) / (
                                        x1 - x0
                                    ) + y0

                                    pixelPos = self.dataToPixel(data, yIntersect)
                                    if pixelPos is not None:
                                        vertices.append((xPixel, yPixel))
                                        vertices.append(pixelPos)
                                    break  # Stop at first intersection

                else:  # y or y2 axes
                    if axis == self.yAxis:
                        axis_name = "left"
                        cornersInData = numpy.array(
                            [self.pixelToData(x, y) for (x, y) in corners]
                        )
                        borders = (
                            (cornersInData[3], cornersInData[2]),  # right
                            (cornersInData[0], cornersInData[3]),  # top
                            (cornersInData[2], cornersInData[1]),
                        )  # bottom

                    else:  # axis == self.y2Axis
                        axis_name = "right"
                        corners = numpy.array(
                            [self.pixelToData(x, y, axis="right") for (x, y) in corners]
                        )
                        borders = (
                            (cornersInData[1], cornersInData[0]),  # left
                            (cornersInData[0], cornersInData[3]),  # top
                            (cornersInData[2], cornersInData[1]),
                        )  # bottom

                    for (xPixel, yPixel), data, text in axis.ticks:
                        if test(text):
                            for (x0, y0), (x1, y1) in borders:
                                if min(y0, y1) <= data < max(y0, y1):
                                    xIntersect = (data - y0) * (x1 - x0) / (
                                        y1 - y0
                                    ) + x0

                                    pixelPos = self.dataToPixel(
                                        xIntersect, data, axis=axis_name
                                    )
                                    if pixelPos is not None:
                                        vertices.append((xPixel, yPixel))
                                        vertices.append(pixelPos)
                                    break  # Stop at first intersection

        return vertices

    def _buildVerticesAndLabels(self):
        width, height = self.size

        xCoords = (self.margins.left - 0.5, width - self.margins.right + 0.5)
        yCoords = (height - self.margins.bottom + 0.5, self.margins.top - 0.5)

        self.axes[0].displayCoords = (
            (xCoords[0], yCoords[0]),
            (xCoords[1], yCoords[0]),
        )

        self._x2AxisCoords = ((xCoords[0], yCoords[1]), (xCoords[1], yCoords[1]))

        # Set order&offset anchor **before** handling Y axis inversion
        fontPixelSize = self._font.pixelSize()
        if fontPixelSize == -1:
            fontPixelSize = self._font.pointSizeF() / 72.0 * self.dotsPerInch

        self.axes[0].orderOffetAnchor = (
            xCoords[1],
            yCoords[0] + fontPixelSize * 1.2,
        )
        self.axes[1].orderOffetAnchor = (
            xCoords[0],
            yCoords[1] - 4 * self.devicePixelRatio,
        )
        self._y2Axis.orderOffetAnchor = (
            xCoords[1],
            yCoords[1] - 4 * self.devicePixelRatio,
        )

        if self.isYAxisInverted:
            # Y axes are inverted, axes coordinates are inverted
            yCoords = yCoords[1], yCoords[0]

        self.axes[1].displayCoords = (
            (xCoords[0], yCoords[0]),
            (xCoords[0], yCoords[1]),
        )

        self._y2Axis.displayCoords = (
            (xCoords[1], yCoords[0]),
            (xCoords[1], yCoords[1]),
        )

        super(GLPlotFrame2D, self)._buildVerticesAndLabels()

        vertices, gridVertices, labels = self._renderResources

        # Adds vertices for borders without axis
        extraVertices = []
        extraVertices += self._x2AxisCoords
        if not self.isY2Axis:
            extraVertices += self._y2Axis.displayCoords

        extraVertices = numpy.asarray(extraVertices, dtype=numpy.float32)
        vertices = numpy.append(vertices, extraVertices, axis=0)

        self._renderResources = (vertices, gridVertices, labels)

    @property
    def foregroundColor(self):
        """Color used for frame and labels"""
        return self._foregroundColor

    @foregroundColor.setter
    def foregroundColor(self, color):
        """Color used for frame and labels"""
        assert len(color) == 4, "foregroundColor must have length 4, got {}".format(
            len(self._foregroundColor)
        )
        if self._foregroundColor != color:
            self._y2Axis.foregroundColor = color
            GLPlotFrame.foregroundColor.fset(self, color)  # call parent property



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2017-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module contains library wide configuration.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "09/11/2018"


class Config(object):
    """
    Class containing shared global configuration for the silx library.

    .. versionadded:: 0.8
    """

    DEFAULT_PLOT_BACKEND = "matplotlib", "opengl"
    """Default plot backend.

    It will be used as default backend for all the next created PlotWidget.

    This attribute can be set with:

    - 'matplotlib' (default) or 'mpl'
    - 'opengl', 'gl'
    - 'none'
    - A :class:`silx.gui.plot.backend.BackendBase.BackendBase` class
    - A callable returning backend class or binding name

    If multiple backends are provided, the first available one is used.

    .. versionadded:: 0.8
    """

    DEFAULT_COLORMAP_NAME = "gray"
    """Default LUT for the plot widgets.

    The available list of names are available in the module
    :module:`silx.gui.colors`.

    .. versionadded:: 0.8
    """

    DEFAULT_PLOT_IMAGE_Y_AXIS_ORIENTATION = "upward"
    """Default Y-axis orientation for plot widget displaying images.

    This attribute can be set with:

    - 'upward' (default), which set the origin to the bottom with an upward
        orientation.
    - 'downward', which set the origin to the top with a backward orientation.

    It will have an influence on:

    - :class:`silx.gui.plot.StackWidget`
    - :class:`silx.gui.plot.ComplexImageView`
    - :class:`silx.gui.plot.Plot2D`
    - :class:`silx.gui.plot.ImageView`

    .. versionadded:: 0.8
    """

    DEFAULT_PLOT_CURVE_COLORS = [
        "#1f77b4",  # tab:blue
        "#ff7f0e",  # tab:orange
        "#2ca02c",  # tab:green
        "#d62728",  # tab:red
        "#9467bd",  # tab:purple
        "#8c564b",  # tab:brown
        "#e377c2",  # tab:pink
        "#7f7f7f",  # tab:gray
        "#bcbd22",  # tab:olive
        "#17becf",  # tab:cyan
    ]
    """Default list of colors for plot widget displaying curves.

    It is based on the color cycle of matplotlib 2.0.
    See https://github.com/matplotlib/matplotlib/blob/main/lib/matplotlib/_color_data.py#L14

    It will have an influence on:

    - :class:`silx.gui.plot.PlotWidget`

    .. versionadded:: 0.9
    """

    DEFAULT_PLOT_CURVE_SYMBOL_MODE = False
    """Whether to display curves with markers or not by default in PlotWidget.

    It will have an influence on PlotWidget curve items.

    .. versionadded:: 0.10
    """

    DEFAULT_PLOT_SYMBOL = "o"
    """Default marker of the item.

    It will have an influence on PlotWidget items

    Supported symbols:

        - 'o', 'Circle'
        - 'd', 'Diamond'
        - 's', 'Square'
        - '+', 'Plus'
        - 'x', 'Cross'
        - '.', 'Point'
        - ',', 'Pixel'
        - '',  'None'

    .. versionadded:: 0.10
    """

    DEFAULT_PLOT_SYMBOL_SIZE = 6.0
    """Default marker size of the item.

    It will have an influence on PlotWidget items

    .. versionadded:: 0.10
    """

    DEFAULT_PLOT_ACTIVE_CURVE_COLOR = None
    """Default color for the active curve.

    It will have an influence on PlotWidget curve items

    .. versionadded:: 2.0
    """

    DEFAULT_PLOT_ACTIVE_CURVE_LINEWIDTH = 2
    """Default line width for the active curve.

    It will have an influence on PlotWidget curve items

    .. versionadded:: 2.0
    """

    DEFAULT_PLOT_MARKER_TEXT_FONT_SIZE = None
    """Default font size for marker text.

    It will have an influence on PlotWidget marker items

    .. versionadded:: 2.0
    """

    _MPL_TIGHT_LAYOUT = False
    """If true the matplotlib backend will use the
    experimental tight layout.

    .. versionadded:: 2.0
    """



#!/usr/bin/python
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Script to update icons.rst file according to icons available in resources.
"""

__authors__ = ["Thomas Vincent"]
__license__ = "MIT"
__date__ = "27/07/2018"


import os
import glob


PROJECT_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

ICONS_RST_DIR = os.path.join(PROJECT_ROOT, "doc", "source", "modules", "gui")

ICONS_RST_FILENAME = os.path.join(ICONS_RST_DIR, "icons.rst")

ICONS_DIR = os.path.join(
    PROJECT_ROOT, "src", "silx", "resources", "gui", "icons", "*.png"
)


ICONS_RST_HEADER = """
.. AUTOMATICALLY GENERATED FILE DO NOT EDIT
   Use %s script instead

.. currentmodule:: silx.gui

:mod:`icons`: Set of icons
--------------------------

.. automodule:: silx.gui.icons
   :members:

Available icons
+++++++++++++++

.. list-table::
   :widths: 1 4
   :header-rows: 1

   * - Icon
     - Name
""" % os.path.basename(
    __file__
)


def main():
    """Write icons.rst file"""
    icons = glob.glob(ICONS_DIR)
    icons = [os.path.relpath(f, ICONS_RST_DIR) for f in icons]
    icons = sorted(icons)

    icons_table = "\n".join(
        "   * - |%s|\n     - %s" % (os.path.basename(f)[:-4], os.path.basename(f)[:-4])
        for f in icons
    )

    icon_definitions = "\n".join(
        ".. |%s| image:: %s" % (os.path.basename(f)[:-4], f) for f in icons
    )

    content = ICONS_RST_HEADER + icons_table + "\n\n" + icon_definitions + "\n"

    # Write to file
    with open(ICONS_RST_FILENAME, "w") as f:
        f.write(content)


if __name__ == "__main__":
    main()



# /*##########################################################################
# Copyright (C) 2016-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/

__authors__ = ["D. Naudet"]
__license__ = "MIT"
__date__ = "01/02/2016"

"""
Tests of the histogramnd function, error cases.
"""
import unittest

import numpy as np

from silx.math.chistogramnd import chistogramnd as histogramnd
from silx.math import Histogramnd


# ==============================================================
# ==============================================================
# ==============================================================


class _Test_chistogramnd_errors(unittest.TestCase):
    """
    Unit tests of the chistogramnd error cases.
    """

    __test__ = False  # ignore abstract class

    def setUp(self):
        self.skipTest("Abstract class")

    def test_weights_shape(self):
        """ """

        for err_w_shape in self.err_weights_shapes:
            test_msg = "Testing invalid weights shape : {0}" "".format(err_w_shape)

            err_weights = np.random.randint(0, high=10, size=err_w_shape)
            err_weights = err_weights.astype(np.double)

            ex_str = None
            try:
                histo, cumul = histogramnd(
                    self.sample, self.histo_range, self.n_bins, weights=err_weights
                )[0:2]
            except ValueError as ex:
                ex_str = str(ex)

            self.assertIsNotNone(ex_str, msg=test_msg)
            self.assertEqual(
                ex_str,
                "<weights> must be an array whose length "
                "is equal to the number of samples.",
            )

    def test_histo_range_shape(self):
        """ """
        n_dims = 1 if len(self.s_shape) == 1 else self.s_shape[1]
        expected_txt_tpl = (
            "<histo_range> error : expected {n_dims} sets "
            "of lower and upper bin edges, "
            "got the following instead : {histo_range}. "
            "(provided <sample> contains "
            "{n_dims}D values)"
        )

        for err_histo_range in self.err_histo_range_shapes:
            test_msg = "Testing invalid histo_range shape : {0}" "".format(
                err_histo_range
            )

            expected_txt = expected_txt_tpl.format(
                histo_range=err_histo_range, n_dims=n_dims
            )

            ex_str = None
            try:
                histo, cumul = histogramnd(
                    self.sample, err_histo_range, self.n_bins, weights=self.weights
                )[0:2]
            except ValueError as ex:
                ex_str = str(ex)

            self.assertIsNotNone(ex_str, msg=test_msg)
            self.assertEqual(ex_str, expected_txt, msg=test_msg)

    def test_nbins_shape(self):
        """ """

        expected_txt = (
            "n_bins must be either a scalar (same number "
            "of bins for all dimensions) or "
            "an array (number of bins for each "
            "dimension)."
        )

        for err_n_bins in self.err_n_bins_shapes:
            test_msg = "Testing invalid n_bins shape : {0}" "".format(err_n_bins)

            ex_str = None
            try:
                histo, cumul = histogramnd(
                    self.sample, self.histo_range, err_n_bins, weights=self.weights
                )[0:2]
            except ValueError as ex:
                ex_str = str(ex)

            self.assertIsNotNone(ex_str, msg=test_msg)
            self.assertEqual(ex_str, expected_txt, msg=test_msg)

    def test_nbins_values(self):
        """ """
        expected_txt = "<n_bins> : only positive values allowed."

        for err_n_bins in self.err_n_bins_values:
            test_msg = "Testing invalid n_bins value : {0}" "".format(err_n_bins)

            ex_str = None
            try:
                histo, cumul = histogramnd(
                    self.sample, self.histo_range, err_n_bins, weights=self.weights
                )[0:2]
            except ValueError as ex:
                ex_str = str(ex)

            self.assertIsNotNone(ex_str, msg=test_msg)
            self.assertEqual(ex_str, expected_txt, msg=test_msg)

    def test_histo_shape(self):
        """ """
        for err_h_shape in self.err_histo_shapes:
            test_msg = "Testing invalid histo shape : {0}" "".format(err_h_shape)

            expected_txt = (
                "Provided <histo> array doesn't have "
                "a shape compatible with <n_bins> "
                ": should be {0} instead of {1}."
                "".format(self.h_shape, err_h_shape)
            )

            histo = np.zeros(shape=err_h_shape, dtype=np.uint32)

            ex_str = None
            try:
                histo, cumul = histogramnd(
                    self.sample,
                    self.histo_range,
                    self.n_bins,
                    weights=self.weights,
                    histo=histo,
                )[0:2]
            except ValueError as ex:
                ex_str = str(ex)

            self.assertIsNotNone(ex_str, msg=test_msg)
            self.assertEqual(ex_str, expected_txt, msg=test_msg)

    def test_histo_dtype(self):
        """ """
        for err_h_dtype in self.err_histo_dtypes:
            test_msg = "Testing invalid histo dtype : {0}" "".format(err_h_dtype)

            histo = np.zeros(shape=self.h_shape, dtype=err_h_dtype)

            expected_txt = (
                "Provided <histo> array doesn't have "
                "the expected type "
                ": should be {0} instead of {1}."
                "".format(np.uint32, histo.dtype)
            )

            ex_str = None
            try:
                histo, cumul = histogramnd(
                    self.sample,
                    self.histo_range,
                    self.n_bins,
                    weights=self.weights,
                    histo=histo,
                )[0:2]
            except ValueError as ex:
                ex_str = str(ex)

            self.assertIsNotNone(ex_str, msg=test_msg)
            self.assertEqual(ex_str, expected_txt, msg=test_msg)

    def test_weighted_histo_shape(self):
        """ """
        # using the same values as histo
        for err_h_shape in self.err_histo_shapes:
            test_msg = "Testing invalid weighted_histo shape : {0}" "".format(
                err_h_shape
            )

            expected_txt = (
                "Provided <weighted_histo> array doesn't have "
                "a shape compatible with <n_bins> "
                ": should be {0} instead of {1}."
                "".format(self.h_shape, err_h_shape)
            )

            cumul = np.zeros(shape=err_h_shape, dtype=np.double)

            ex_str = None
            try:
                histo, cumul = histogramnd(
                    self.sample,
                    self.histo_range,
                    self.n_bins,
                    weights=self.weights,
                    weighted_histo=cumul,
                )[0:2]
            except ValueError as ex:
                ex_str = str(ex)

            self.assertIsNotNone(ex_str, msg=test_msg)
            self.assertEqual(ex_str, expected_txt, msg=test_msg)

    def test_cumul_dtype(self):
        """ """
        # using the same values as histo
        for err_h_dtype in self.err_histo_dtypes:
            test_msg = "Testing invalid weighted_histo dtype : {0}" "".format(
                err_h_dtype
            )

            cumul = np.zeros(shape=self.h_shape, dtype=err_h_dtype)

            expected_txt = (
                "Provided <weighted_histo> array doesn't have "
                "the expected type "
                ": should be {0} or {1} instead of {2}."
                "".format(np.float64, np.float32, cumul.dtype)
            )

            ex_str = None
            try:
                histo, cumul = histogramnd(
                    self.sample,
                    self.histo_range,
                    self.n_bins,
                    weights=self.weights,
                    weighted_histo=cumul,
                )[0:2]
            except ValueError as ex:
                ex_str = str(ex)

            self.assertIsNotNone(ex_str, msg=test_msg)
            self.assertEqual(ex_str, expected_txt, msg=test_msg)

    def test_wh_histo_dtype(self):
        """ """
        # using the same values as histo
        for err_h_dtype in self.err_histo_dtypes:
            test_msg = "Testing invalid wh_dtype dtype : {0}" "".format(err_h_dtype)

            expected_txt = "<wh_dtype> type not supported : {0}." "".format(err_h_dtype)

            ex_str = None
            try:
                histo, cumul = histogramnd(
                    self.sample,
                    self.histo_range,
                    self.n_bins,
                    weights=self.weights,
                    wh_dtype=err_h_dtype,
                )[0:2]
            except ValueError as ex:
                ex_str = str(ex)

            self.assertIsNotNone(ex_str, msg=test_msg)
            self.assertEqual(ex_str, expected_txt, msg=test_msg)

    def test_unmanaged_dtypes(self):
        """ """
        for err_unmanaged_dtype in self.err_unmanaged_dtypes:
            test_msg = "Testing unmanaged dtypes : {0}" "".format(err_unmanaged_dtype)

            sample = self.sample.astype(err_unmanaged_dtype[0])
            weights = self.weights.astype(err_unmanaged_dtype[1])

            expected_txt = (
                "Case not supported - sample:{0} "
                "and weights:{1}."
                "".format(sample.dtype, weights.dtype)
            )

            ex_str = None
            try:
                histogramnd(sample, self.histo_range, self.n_bins, weights=weights)
            except TypeError as ex:
                ex_str = str(ex)

            self.assertIsNotNone(ex_str, msg=test_msg)
            self.assertEqual(ex_str, expected_txt, msg=test_msg)

    def test_uncontiguous_histo(self):
        """ """
        # non contiguous array
        shape = np.array(self.n_bins, ndmin=1)
        shape[0] *= 2
        histo_tmp = np.zeros(shape)
        histo = histo_tmp[::2, ...]

        expected_txt = "<histo> must be a C_CONTIGUOUS numpy array."

        ex_str = None
        try:
            histogramnd(
                self.sample,
                self.histo_range,
                self.n_bins,
                weights=self.weights,
                histo=histo,
            )
        except ValueError as ex:
            ex_str = str(ex)

        self.assertIsNotNone(ex_str)
        self.assertEqual(ex_str, expected_txt)

    def test_uncontiguous_weighted_histo(self):
        """ """
        # non contiguous array
        shape = np.array(self.n_bins, ndmin=1)
        shape[0] *= 2
        cumul_tmp = np.zeros(shape)
        cumul = cumul_tmp[::2, ...]

        expected_txt = "<weighted_histo> must be a C_CONTIGUOUS numpy array."

        ex_str = None
        try:
            histogramnd(
                self.sample,
                self.histo_range,
                self.n_bins,
                weights=self.weights,
                weighted_histo=cumul,
            )
        except ValueError as ex:
            ex_str = str(ex)

        self.assertIsNotNone(ex_str)
        self.assertEqual(ex_str, expected_txt)


class Test_chistogramnd_1D_errors(_Test_chistogramnd_errors):
    """
    Unit tests of the 1D histogramnd error cases.
    """

    __test__ = True  # because _Test_chistogramnd_errors is ignored

    def setUp(self):
        # nominal values
        self.n_elements = 1000
        self.s_shape = (self.n_elements,)
        self.w_shape = (self.n_elements,)

        self.histo_range = [0.0, 100.0]
        self.n_bins = 10

        self.h_shape = (self.n_bins,)

        self.sample = np.random.randint(0, high=10, size=self.s_shape)
        self.sample = self.sample.astype(np.double)

        self.weights = np.random.randint(0, high=10, size=self.w_shape)
        self.weights = self.weights.astype(np.double)

        self.err_weights_shapes = (
            (self.n_elements + 1,),
            (self.n_elements - 1,),
            (self.n_elements - 1, 3),
        )
        self.err_histo_range_shapes = ([0.0], [0.0, 1.0, 2.0], [[0.0], [1.0]])
        self.err_n_bins_shapes = ([10, 2], [[10], [2]])
        self.err_n_bins_values = (0, [-10], None)
        self.err_histo_shapes = (
            (self.n_bins + 1,),
            (self.n_bins - 1,),
            (self.n_bins, self.n_bins),
        )
        # these are used for testing the histo parameter as well
        #   as the weighted_histo parameter.
        self.err_histo_dtypes = (np.uint16, np.float16)

        self.err_unmanaged_dtypes = (
            (np.double, np.uint16),
            (np.uint16, np.double),
            (np.uint16, np.uint16),
        )


class Test_chistogramnd_ND_range(unittest.TestCase):
    """ """

    def test_invalid_histo_range(self):
        data = np.random.random((60, 60))
        nbins = 10

        with self.assertRaises(ValueError):
            histo_range = data.min(), np.inf

            Histogramnd(sample=data.ravel(), histo_range=histo_range, n_bins=nbins)

            histo_range = data.min(), np.nan

            Histogramnd(sample=data.ravel(), histo_range=histo_range, n_bins=nbins)


class Test_chistogramnd_ND_errors(_Test_chistogramnd_errors):
    """
    Unit tests of the 3D histogramnd error cases.
    """

    __test__ = True  # because _Test_chistogramnd_errors is ignored

    def setUp(self):
        # nominal values
        self.n_elements = 1000
        self.s_shape = (self.n_elements, 3)
        self.w_shape = (self.n_elements,)

        self.histo_range = [[0.0, 100.0], [0.0, 100.0], [0.0, 100.0]]
        self.n_bins = (10, 20, 30)

        self.h_shape = self.n_bins

        self.sample = np.random.randint(0, high=10, size=self.s_shape)
        self.sample = self.sample.astype(np.double)

        self.weights = np.random.randint(0, high=10, size=self.w_shape)
        self.weights = self.weights.astype(np.double)

        self.err_weights_shapes = (
            (self.n_elements + 1,),
            (self.n_elements - 1,),
            (self.n_elements - 1, 3),
        )
        self.err_histo_range_shapes = (
            [0.0],
            [0.0, 1.0],
            [[0.0, 10.0], [0.0, 10.0]],
            [0.0, 10.0, 0, 10.0, 0, 10.0],
        )
        self.err_n_bins_shapes = ([10, 2], [[10], [20], [30]])
        self.err_n_bins_values = (0, [-10], [10, 20, -4], None, [10, None, 30])
        self.err_histo_shapes = (
            (self.n_bins[0] + 1, self.n_bins[1], self.n_bins[2]),
            (self.n_bins[0], self.n_bins[1], self.n_bins[2] - 1),
            (self.n_bins[0], self.n_bins[1]),
            (self.n_bins[1], self.n_bins[0], self.n_bins[2]),
            (self.n_bins[0], self.n_bins[1], self.n_bins[2], 10),
        )
        # these are used for testing the histo parameter as well
        #   as the weighted_histo parameter.
        self.err_histo_dtypes = (np.uint16, np.float16)

        self.err_unmanaged_dtypes = (
            (np.double, np.uint16),
            (np.uint16, np.double),
            (np.uint16, np.uint16),
        )
